from immutabledict import immutabledict
import strax
import straxen
from copy import deepcopy
from straxen import HAVE_ADMIX
import os
import warnings
import typing as ty
from pandas.util._decorators import deprecate_kwarg
import socket

common_opts = dict(
    register_all=[],
    # Register all peak/pulse processing by hand as 1T does not need to have
    # the high-energy plugins.
    register=[
        straxen.PulseProcessing,
        straxen.Peaklets,
        straxen.PeakletClassification,
        straxen.MergedS2s,
        straxen.Peaks,
        straxen.PeakBasics,
        straxen.PeakProximity,
        straxen.Events,
        straxen.EventBasics,
        straxen.EventPositions,
        straxen.CorrectedAreas,
        straxen.EnergyEstimates,
        straxen.EventInfoDouble,
        straxen.DistinctChannels,
    ],
    check_available=('raw_records', 'peak_basics'),
    store_run_fields=(
        'name', 'number',
        'start', 'end', 'livetime', 'mode', 'source'))

xnt_common_config = dict(
    n_tpc_pmts=straxen.n_tpc_pmts,
    n_top_pmts=straxen.n_top_pmts,
    gain_model="cmt://to_pe_model?version=ONLINE&run_id=plugin.run_id",
    gain_model_nv="cmt://to_pe_model_nv?version=ONLINE&run_id=plugin.run_id",
    gain_model_mv="cmt://to_pe_model_mv?version=ONLINE&run_id=plugin.run_id",
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        # Channels must be listed in a ascending order!
        tpc=(0, 493),
        he=(500, 752),  # high energy
        aqmon=(790, 807),
        aqmon_nv=(808, 815),  # nveto acquisition monitor
        tpc_blank=(999, 999),
        mv=(1000, 1083),
        aux_mv=(1084, 1087),  # Aux mv channel 2 empty  1 pulser  and 1 GPS
        mv_blank=(1999, 1999),
        nveto=(2000, 2119),
        nveto_blank=(2999, 2999)),
    # Clustering/classification parameters
    # Event level parameters
    fdc_map='itp_map://'
            'resource://'
            'cmt://'
            'format://fdc_map_{alg}'
            '?alg=plugin.default_reconstruction_algorithm'
            '&version=ONLINE'
            '&run_id=plugin.run_id'
            '&fmt=binary'
            '&scale_coordinates=plugin.coordinate_scales',
    z_bias_map='itp_map://'
               'resource://'
               'XnT_z_bias_map_chargeup_20230329.json.gz?'
               'fmt=json.gz'
               '&method=RegularGridInterpolator',
)
# these are placeholders to avoid calling cmt with non integer run_ids. Better solution pending.
# s1,s2 and fd corrections are still problematic
xnt_simulation_config = deepcopy(xnt_common_config)
xnt_simulation_config.update(gain_model="legacy-to-pe://to_pe_placeholder",
                             gain_model_nv="legacy-to-pe://adc_nv",
                             gain_model_mv="legacy-to-pe://adc_mv",
                             elife=1e6,
                             )

# Plugins in these files have nT plugins, E.g. in pulse&peak(let)
# processing there are plugins for High Energy plugins. Therefore, do not
# st.register_all in 1T contexts.
xnt_common_opts = common_opts.copy()
xnt_common_opts.update({
    'register': common_opts['register'] + [],
    'register_all': common_opts['register_all'] + [straxen.plugins,
                                                   ],
    'use_per_run_defaults': False,
})


##
# XENONnT
##


def xenonnt(cmt_version='global_ONLINE', xedocs_version=None,
            _from_cutax=False, **kwargs):
    """XENONnT context"""
    if not _from_cutax and cmt_version != 'global_ONLINE':
        warnings.warn('Don\'t load a context directly from straxen, '
                      'use cutax instead!')
    st = straxen.contexts.xenonnt_online(**kwargs)
    st.apply_cmt_version(cmt_version)

    if xedocs_version is not None:
        st.apply_xedocs_configs(version=xedocs_version, **kwargs)

    return st


def find_rucio_local_path(include_rucio_local, _rucio_local_path):
    """
    Check the hostname to determine which rucio local path to use. Note that access to
    /dali/lgrandi/rucio/ is possible only if you are on dali compute node or login node.

    :param include_rucio_local: add the rucio local storage frontend.
        This is only needed if one wants to do a fuzzy search in the
        data the runs database is out of sync with rucio
    :param _rucio_local_path: str, path of local RSE of rucio. Only use
        for testing!
    """
    hostname = socket.gethostname()
    # if you are on dali compute node, do nothing
    if ('dali' in hostname) and ('login' not in hostname):
        _include_rucio_local = include_rucio_local
        __rucio_local_path = _rucio_local_path
    # Assumed the only other option is 'midway' or login nodes, 
    # where we have full access to dali and project space. 
    # This doesn't make sense outside XENON but we don't care.
    else:
        _include_rucio_local = True
        __rucio_local_path = '/project/lgrandi/rucio/'
        print('You specified _auto_append_rucio_local=True and you are not on dali compute nodes,'
              'so we will add the following rucio local path: ', __rucio_local_path)

    return _include_rucio_local, __rucio_local_path


@deprecate_kwarg('_minimum_run_number', 'minimum_run_number')
@deprecate_kwarg('_maximum_run_number', 'maximum_run_number')
@deprecate_kwarg('_include_rucio_remote', 'include_rucio_remote')
@deprecate_kwarg('_add_online_monitor_frontend', 'include_online_monitor')
def xenonnt_online(output_folder: str = './strax_data',
                   we_are_the_daq: bool = False,
                   minimum_run_number: int = 7157,
                   maximum_run_number: ty.Optional[int] = None,

                   # Frontends
                   include_rucio_remote: bool = False,
                   include_online_monitor: bool = False,
                   include_rucio_local: bool = False,

                   # Frontend options
                   download_heavy: bool = False,
                   _auto_append_rucio_local: bool = True,
                   _rucio_path: str = '/dali/lgrandi/rucio/',
                   _rucio_local_path: ty.Optional[str] = None,
                   _raw_paths: ty.Optional[str] = ['/dali/lgrandi/xenonnt/raw'],
                   _processed_paths: ty.Optional[ty.List[str]] = ['/dali/lgrandi/xenonnt/processed',
                                                                  '/project2/lgrandi/xenonnt/processed',
                                                                  '/project/lgrandi/xenonnt/processed'],

                   # Testing options
                   _context_config_overwrite: ty.Optional[dict] = None,
                   _database_init: bool = True,
                   _forbid_creation_of: ty.Optional[dict] = None,
                   **kwargs):
    """
    XENONnT online processing and analysis

    :param output_folder: str, Path of the strax.DataDirectory where new
        data can be stored
    :param we_are_the_daq: bool, if we have admin access to upload data
    :param minimum_run_number: int, lowest number to consider
    :param maximum_run_number: Highest number to consider. When None
        (the default) consider all runs that are higher than the
        minimum_run_number.
    :param include_rucio_remote: add the rucio remote frontend to the
        context
    :param include_online_monitor: add the online monitor storage frontend.
    :param include_rucio_local: add the rucio local storage frontend.
        This is only needed if one wants to do a fuzzy search in the
        data the runs database is out of sync with rucio
    :param download_heavy: bool, whether or not to allow downloads of
        heavy data (raw_records*, less the aqmon)

    :param _auto_append_rucio_local: bool, whether or not to automatically append the 
        rucio local path
    :param _rucio_path: str, path of rucio
    :param _rucio_local_path: str, path of local RSE of rucio. Only use
        for testing!
    :param _raw_paths: list[str], common path of the raw-data
    :param _processed_paths: list[str]. common paths of output data
    :param _context_config_overwrite: dict, overwrite config
    :param _database_init: bool, start the database (for testing)
    :param _forbid_creation_of: str/tuple, of datatypes to prevent form
        being written (raw_records* is always forbidden).
    :param kwargs: dict, context options
    :return: strax.Context
    """
    context_options = {
        **straxen.contexts.xnt_common_opts,
        **kwargs}

    st = strax.Context(
        config=straxen.contexts.xnt_common_config,
        **context_options)
    st.register([straxen.DAQReader,
                 straxen.LEDCalibration,
                 straxen.LEDAfterpulseProcessing])

    if _auto_append_rucio_local:
        include_rucio_local, _rucio_local_path = find_rucio_local_path(
            include_rucio_local, _rucio_local_path)

    st.storage = [
        straxen.RunDB(
            readonly=not we_are_the_daq,
            minimum_run_number=minimum_run_number,
            maximum_run_number=maximum_run_number,
            runid_field='number',
            new_data_path=output_folder,
            rucio_path=_rucio_path,
        )] if _database_init else []
    if not we_are_the_daq:
        for _raw_path in _raw_paths:
            st.storage += [
                strax.DataDirectory(
                    _raw_path,
                    readonly=True,
                    take_only=straxen.DAQReader.provides)]
        for _processed_path in _processed_paths:
            st.storage += [strax.DataDirectory(
                _processed_path,
                readonly=True)]

        if output_folder:
            st.storage += [strax.DataDirectory(output_folder,
                                               provide_run_metadata=True,
                                               )]
        st.context_config['forbid_creation_of'] = straxen.daqreader.DAQReader.provides
        if _forbid_creation_of is not None:
            st.context_config['forbid_creation_of'] += strax.to_str_tuple(_forbid_creation_of)

    # Add the rucio frontend if we are able to
    if include_rucio_remote and HAVE_ADMIX:
        rucio_frontend = straxen.RucioRemoteFrontend(
            staging_dir=os.path.join(output_folder, 'rucio'),
            download_heavy=download_heavy,
        )
        st.storage += [rucio_frontend]

    if include_rucio_local:
        rucio_local_frontend = straxen.RucioLocalFrontend(path=_rucio_local_path)
        st.storage += [rucio_local_frontend]

    # Only the online monitor backend for the DAQ
    if _database_init and (include_online_monitor or we_are_the_daq):
        st.storage += [straxen.OnlineMonitor(
            readonly=not we_are_the_daq,
            take_only=('veto_intervals',
                       'online_peak_monitor',
                       'event_basics',
                       'online_monitor_nv',
                       'online_monitor_mv',
                       'individual_peak_monitor',
                       ))]

    # Remap the data if it is before channel swap (because of wrongly cabled
    # signal cable connectors) These are runs older than run 8797. Runs
    # newer than 8796 are not affected. See:
    # https://github.com/XENONnT/straxen/pull/166 and
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:daq:sector_swap
    st.set_context_config({'apply_data_function': (straxen.remap_old,
                                                   straxen.check_loading_allowed,
                                                   )})
    if _context_config_overwrite is not None:
        warnings.warn(f'_context_config_overwrite is deprecated, please pass to context as kwargs',
                      DeprecationWarning)
        st.set_context_config(_context_config_overwrite)

    return st


def xenonnt_led(**kwargs):
    st = xenonnt_online(**kwargs)
    st.set_context_config(
        {'check_available': ('raw_records', 'led_calibration'),
         'free_options': list(xnt_common_config.keys())
         })
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.DAQReader,
                 straxen.LEDCalibration,
                 straxen.nVETORecorder,
                 straxen.nVETOPulseProcessing,
                 straxen.nVETOHitlets,
                 straxen.nVetoExtTimings, ])
    st.set_config({"coincidence_level_recorder_nv": 1})
    return st


def xenonnt_simulation_offline(output_folder: str = './strax_data',
                               wfsim_registry: str = 'RawRecordsFromFaxNT',
                               run_id: ty.Optional[str] = None,
                               global_version: ty.Optional[str] = None,
                               fax_config: ty.Optional[str] = None,
                               ):
    """
    :param output_folder: strax_data folder
    :param wfsim_registry: Raw_records generation mechanism,
                           'RawRecordsFromFaxNT', 'RawRecordsFromMcChain', etc,
                           https://github.com/XENONnT/WFSim/blob/master/wfsim/strax_interface.py
    :param run_id: Real run_id to use to fetch the corrections
    :param global_version: Global versions
                           https://github.com/XENONnT/corrections/tree/master/XENONnT/global_versions
    :param fax_config: WFSim configuration files
                       https://github.com/XENONnT/private_nt_aux_files/blob/master/sim_files/fax_config_nt_sr0_v4.json
    :return: strax context for simulation
    """
    if run_id is None:
        raise ValueError("Specify a run_id to load the corrections")
    if global_version is None:
        raise ValueError("Specify a correction global version")
    if fax_config is None:
        raise ValueError("Specify a simulation configuration file")

    import wfsim
    # General strax context, register common plugins
    st = strax.Context(storage=strax.DataDirectory(output_folder),
                       **straxen.contexts.xnt_common_opts)
    # Register simulation configs required by WFSim plugins
    st.config.update(dict(detector='XENONnT',
                          fax_config=fax_config,
                          check_raw_record_overlaps=True,
                          **straxen.contexts.xnt_common_config))
    # Register WFSim raw_records plugin to overwrite real data raw_records
    wfsim_plugin = getattr(wfsim, wfsim_registry)
    st.register(wfsim_plugin)
    for plugin_name in wfsim_plugin.provides:
        assert 'wfsim' in str(st._plugin_class_registry[plugin_name])
    # Register offline global corrections same as real data
    st.apply_xedocs_configs(version=global_version)
    # Real data correction is run_id dependent,
    # but in sim we often use run_id not in the runDB
    # So we switch the run_id dependence to a specific run -> run_id
    local_versions = st.config
    for config_name, url_config in local_versions.items():
        if isinstance(url_config, str):
            if 'run_id' in url_config:
                local_versions[config_name] = straxen.URLConfig.format_url_kwargs(url_config, run_id=run_id)
    st.config = local_versions
    # In simulation, the raw_records generation depends on gain measurement
    st.config['gain_model_mc'] = st.config['gain_model']
    # No blinding in simulations
    st.config["event_info_function"] = "disabled"
    return st


def xenonnt_simulation(
        output_folder='./strax_data',
        wfsim_registry='RawRecordsFromFaxNT',
        cmt_run_id_sim=None,
        cmt_run_id_proc=None,
        cmt_version='global_ONLINE',
        fax_config='fax_config_nt_design.json',
        overwrite_from_fax_file_sim=False,
        overwrite_from_fax_file_proc=False,
        cmt_option_overwrite_sim=immutabledict(),
        cmt_option_overwrite_proc=immutabledict(),
        _forbid_creation_of=None,
        _config_overlap=immutabledict(
            drift_time_gate='electron_drift_time_gate',
            drift_velocity_liquid='electron_drift_velocity',
            electron_lifetime_liquid='elife',
        ),
        **kwargs):
    """
    The most generic context that allows for setting full divergent
    settings for simulation purposes

    It makes full divergent setup, allowing to set detector simulation
    part (i.e. for wfsim up to truth and  raw_records). Parameters _sim
    refer to detector simulation parameters.

    Arguments having _proc in their name refer to detector parameters that
    are used for processing of simulations as done to the real detector
    data. This means starting from already existing raw_records and finishing
    with higher level data, such as peaks, events etc.

    If only one cmt_run_id is given, the second one will be set automatically,
    resulting in CMT match between simulation and processing. However, detector
    parameters can be still overwritten from fax file or manually using cmt
    config overwrite options.

    CMT options can also be overwritten via fax config file.
    :param output_folder: Output folder for strax data.
    :param wfsim_registry: Name of WFSim plugin used to generate data.
    :param cmt_run_id_sim: Run id for detector parameters from CMT to be used
        for creation of raw_records.
    :param cmt_run_id_proc: Run id for detector parameters from CMT to be used
        for processing from raw_records to higher level data.
    :param cmt_version: Global version for corrections to be loaded.
    :param fax_config: Fax config file to use.
    :param overwrite_from_fax_file_sim: If true sets detector simulation
        parameters for truth/raw_records from from fax_config file istead of CMT
    :param overwrite_from_fax_file_proc:  If true sets detector processing
        parameters after raw_records(peaklets/events/etc) from from fax_config
        file instead of CMT
    :param cmt_option_overwrite_sim: Dictionary to overwrite CMT settings for
        the detector simulation part.
    :param cmt_option_overwrite_proc: Dictionary to overwrite CMT settings for
        the data processing part.
    :param _forbid_creation_of: str/tuple, of datatypes to prevent form
        being written (e.g. 'raw_records' for read only simulation context).
    :param _config_overlap: Dictionary of options to overwrite. Keys
        must be simulation config keys, values must be valid CMT option keys.
    :param kwargs: Additional kwargs taken by strax.Context.
    :return: strax.Context instance
    """
    import wfsim
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(detector='XENONnT',
                    fax_config=fax_config,
                    check_raw_record_overlaps=True,
                    **straxen.contexts.xnt_common_config, ),
        **straxen.contexts.xnt_common_opts, **kwargs)
    st.register(getattr(wfsim, wfsim_registry))

    # Make sure that the non-simulated raw-record types are not requested
    st.deregister_plugins_with_missing_dependencies()

    if straxen.utilix_is_configured(
            warning_message='Bad context as we cannot set CMT since we '
                            'have no database access'''):
        st.apply_cmt_version(cmt_version)

    if _forbid_creation_of is not None:
        st.context_config['forbid_creation_of'] += strax.to_str_tuple(_forbid_creation_of)

    # doing sanity checks for cmt run ids for simulation and processing
    if (not cmt_run_id_sim) and (not cmt_run_id_proc):
        raise RuntimeError("cmt_run_id_sim and cmt_run_id_proc are None. "
                           "You have to specify at least one CMT run id. ")
    if (cmt_run_id_sim and cmt_run_id_proc) and (cmt_run_id_sim != cmt_run_id_proc):
        print("INFO : divergent CMT runs for simulation and processing")
        print("    cmt_run_id_sim".ljust(25), cmt_run_id_sim)
        print("    cmt_run_id_proc".ljust(25), cmt_run_id_proc)
    else:
        cmt_id = cmt_run_id_sim or cmt_run_id_proc
        cmt_run_id_sim = cmt_id
        cmt_run_id_proc = cmt_id

    # Replace default cmt options with cmt_run_id tag + cmt run id
    cmt_options_full = straxen.get_corrections.get_cmt_options(st)

    # prune to just get the strax options
    cmt_options = {key: val['strax_option']
                   for key, val in cmt_options_full.items()}

    # First, fix gain model for simulation
    st.set_config({'gain_model_mc':
                       ('cmt_run_id', cmt_run_id_sim, *cmt_options['gain_model'])})
    fax_config_override_from_cmt = dict()
    for fax_field, cmt_field in _config_overlap.items():
        value = cmt_options[cmt_field]

        # URL configs need to be converted to the expected format
        if isinstance(value, str):
            opt_cfg = cmt_options_full[cmt_field]
            version = straxen.URLConfig.kwarg_from_url(value, 'version')
            # We now allow the cmt name to be different from the config name
            # WFSim expects the cmt name
            value = (opt_cfg['correction'], version, True)

        fax_config_override_from_cmt[fax_field] = ('cmt_run_id', cmt_run_id_sim,
                                                   *value)
    st.set_config({'fax_config_override_from_cmt': fax_config_override_from_cmt})

    # and all other parameters for processing
    for option in cmt_options:
        value = cmt_options[option]
        if isinstance(value, str):
            # for URL configs we can just replace the run_id keyword argument
            # This will become the proper way to override the run_id for cmt configs
            st.config[option] = straxen.URLConfig.format_url_kwargs(value, run_id=cmt_run_id_proc)
        else:
            # FIXME: Remove once all cmt configs are URLConfigs
            st.config[option] = ('cmt_run_id', cmt_run_id_proc, *value)

    # Done with "default" usage, now to overwrites from file
    #
    # Take fax config and put into context option
    if overwrite_from_fax_file_proc or overwrite_from_fax_file_sim:
        fax_config = straxen.get_resource(fax_config, fmt='json')
        for fax_field, cmt_field in _config_overlap.items():
            if overwrite_from_fax_file_proc:
                if isinstance(cmt_options[cmt_field], str):
                    # URLConfigs can just be set to a constant
                    st.config[cmt_field] = fax_config[fax_field]
                else:
                    # FIXME: Remove once all cmt configs are URLConfigs
                    st.config[cmt_field] = (cmt_options[cmt_field][0] + '_constant',
                                            fax_config[fax_field])
            if overwrite_from_fax_file_sim:
                # CMT name allowed to be different from the config name
                # WFSim needs the cmt name
                cmt_name = cmt_options_full[cmt_field]['correction']

                st.config['fax_config_override_from_cmt'][fax_field] = (
                    cmt_name + '_constant', fax_config[fax_field])

    # And as the last step - manual overrrides, since they have the highest priority
    # User customized for simulation
    for option in cmt_option_overwrite_sim:
        if option not in cmt_options:
            raise ValueError(f'Overwrite option {option} is not using CMT by default '
                             'you should just use set config')
        if option not in _config_overlap.values():
            raise ValueError(f'Overwrite option {option} does not have mapping from '
                             f'CMT to fax config!')
        for fax_key, cmt_key in _config_overlap.items():
            if cmt_key == option:
                cmt_name = cmt_options_full[option]['correction']
                st.config['fax_config_override_from_cmt'][fax_key] = (
                    cmt_name + '_constant',
                    cmt_option_overwrite_sim[option])
            del (fax_key, cmt_key)
    # User customized for simulation
    for option in cmt_option_overwrite_proc:
        if option not in cmt_options:
            raise ValueError(f'Overwrite option {option} is not using CMT by default '
                             'you should just use set config')

        if isinstance(cmt_options[option], str):
            # URLConfig options can just be set to constants, no hacks needed
            # But for now lets keep things consistent for people
            st.config[option] = cmt_option_overwrite_proc[option]
        else:
            # CMT name allowed to be different from the config name
            # WFSim needs the cmt name
            cmt_name = cmt_options_full[option]['correction']
            st.config[option] = (cmt_name + '_constant',
                                 cmt_option_overwrite_proc[option])
    # Only for simulations
    st.set_config({"event_info_function": "disabled"})

    return st


##
# XENON1T, see straxen/legacy
##


def demo():
    """Return strax context used in the straxen demo notebook"""
    return straxen.legacy.contexts_1t.demo()


def fake_daq():
    """Context for processing fake DAQ data in the current directory"""
    return straxen.legacy.contexts_1t.fake_daq()


def xenon1t_dali(output_folder='./strax_data', build_lowlevel=False, **kwargs):
    return straxen.legacy.contexts_1t.xenon1t_dali(output_folder=output_folder, build_lowlevel=build_lowlevel, **kwargs)


def xenon1t_led(**kwargs):
    return straxen.legacy.contexts_1t.xenon1t_led(**kwargs)


def xenon1t_simulation(output_folder='./strax_data'):
    return straxen.legacy.contexts_1t.xenon1t_simulation(output_folder=output_folder)



import strax
import straxen
import numpy as np
import unittest
from strax.testutils import run_id
from hypothesis import strategies, given, settings

TEST_DATA_LENGTH = 3
R_TOL_DEFAULT = 1e-5


def _not_close_to_0_or_1(x, rtol=R_TOL_DEFAULT):
    return not (np.isclose(x, 1, rtol=rtol) or np.isclose(x, 0, rtol=rtol))


class TestComputePeakBasics(unittest.TestCase):
    """Tests for peak basics plugin"""
    def setUp(self, context=straxen.contexts.demo):
        self.st = context()
        self.n_top = self.st.config.get('n_top_pmts', 2)

        # Make sure that the check is on. Otherwise we cannot test it.
        self.st.set_config({'check_peak_sum_area_rtol': R_TOL_DEFAULT})
        self.peaks_basics_compute = self.st.get_single_plugin(run_id, 'peak_basics').compute

    @settings(deadline=None)
    @given(strategies.integers(min_value=0,
                               max_value=TEST_DATA_LENGTH - 1),
           )
    def test_aft_equals1(self, test_peak_idx):
        """Fill top array with area 1"""
        test_data = self.get_test_peaks(self.n_top)
        test_data[test_peak_idx]['area_per_channel'][:self.n_top] = 1
        test_data[test_peak_idx]['area'] = np.sum(test_data[test_peak_idx]['area_per_channel'])
        peaks = self.peaks_basics_compute(test_data)
        assert peaks[test_peak_idx]['area_fraction_top'] == 1

    @settings(deadline=None)
    @given(strategies.floats(min_value=0,
                             max_value=2,
                             ).filter(_not_close_to_0_or_1),
           strategies.integers(min_value=0,
                               max_value=TEST_DATA_LENGTH - 1,
                               ),
           )
    def test_bad_peak(self, off_by_factor, test_peak_idx):
        """
        Lets deliberately make some data that is not self-consistent to
            run into the error in the test.
        """
        test_data = self.get_test_peaks(self.n_top)
        test_data[test_peak_idx]['area_per_channel'][:self.n_top] = 1
        area = np.sum(test_data[test_peak_idx]['area_per_channel'])

        # Set the area field to a wrong value
        area *= off_by_factor
        test_data[test_peak_idx]['area'] = area
        self.assertRaises(ValueError, self.peaks_basics_compute, test_data)

    @staticmethod
    def get_test_peaks(n_top, length=2, sum_wf_samples=10):
        """Generate some dummy peaks"""
        test_data = np.zeros(TEST_DATA_LENGTH,
                             dtype=strax.dtypes.peak_dtype(
                                 n_channels=n_top + 1,
                                 n_sum_wv_samples=sum_wf_samples)
                             )
        test_data['time'] = range(TEST_DATA_LENGTH)
        test_data['time'] *= length * 2
        test_data['dt'] = 1
        test_data['length'] = length
        return test_data



import numpy as np
import strax
import straxen

from straxen.plugins.defaults import FAKE_MERGED_S2_TYPE

export, __all__ = strax.exporter()


@export
class Peaks(strax.Plugin):
    """
    Merge peaklets and merged S2s such that we obtain our peaks
    (replacing all peaklets that were later re-merged as S2s). As this
    step is computationally trivial, never save this plugin.
    """
    __version__ = '0.1.2'

    depends_on = ('peaklets', 'peaklet_classification', 'merged_s2s')
    data_kind = 'peaks'
    provides = 'peaks'
    parallel = True
    compressor = 'zstd'
    save_when = strax.SaveWhen.EXPLICIT

    diagnose_sorting = straxen.URLConfig(
        track=False, default=False, infer_type=False,
        help="Enable runtime checks for sorting and disjointness")

    merge_without_s1 = straxen.URLConfig(
        default=True, infer_type=False,
        help="If true, S1s will be igored during the merging. "
             "It's now possible for a S1 to be inside a S2 post merging")

    def infer_dtype(self):
        return self.deps['peaklets'].dtype_for('peaklets')

    def compute(self, peaklets, merged_s2s):
        # Remove fake merged S2s from dirty hack, see above
        merged_s2s = merged_s2s[merged_s2s['type'] != FAKE_MERGED_S2_TYPE]

        if self.merge_without_s1:
            is_s1 = peaklets['type'] == 1
            peaks = strax.replace_merged(peaklets[~is_s1], merged_s2s)
            peaks = strax.sort_by_time(np.concatenate([peaklets[is_s1],
                                                       peaks]))
        else:
            peaks = strax.replace_merged(peaklets, merged_s2s)

        if self.diagnose_sorting:
            assert np.all(np.diff(peaks['time']) >= 0), "Peaks not sorted"
            if self.merge_without_s1:
                to_check = peaks['type'] != 1
            else:
                to_check = peaks['type'] != FAKE_MERGED_S2_TYPE

            assert np.all(peaks['time'][to_check][1:]
                          >= strax.endtime(peaks)[to_check][:-1]), "Peaks not disjoint"
        return peaks



from immutabledict import immutabledict
import strax
import straxen

export, __all__ = strax.exporter()

from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.records.records import PulseProcessing, pulse_count_dtype


@export
class PulseProcessingHighEnergy(PulseProcessing):
    __doc__ = HE_PREAMBLE + PulseProcessing.__doc__
    __version__ = '0.0.1'
    provides = ('records_he', 'pulse_counts_he')
    data_kind = {k: k for k in provides}
    rechunk_on_save = immutabledict(
        records_he=False,
        pulse_counts_he=True)
    depends_on = 'raw_records_he'
    compressor = 'zstd'
    child_plugin = True
    save_when = strax.SaveWhen.TARGET

    n_he_pmts = straxen.URLConfig(
        track=False, default=752, infer_type=False,
        help="Maximum channel of the he channels")

    record_length = straxen.URLConfig(
        default=110, track=False, type=int,
        help="Number of samples per raw_record")

    hit_min_amplitude_he = straxen.URLConfig(
        default='cmt://hit_thresholds_he?version=ONLINE&run_id=plugin.run_id', track=True, infer_type=False,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_tpc_pmts, or a number,'
             'or a string like "pmt_commissioning_initial" which means calling'
             'hitfinder_thresholds.py'
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.'
    )

    def infer_dtype(self):
        dtype = dict()
        dtype['records_he'] = strax.record_dtype(self.record_length)
        dtype['pulse_counts_he'] = pulse_count_dtype(self.n_he_pmts)
        return dtype

    def setup(self):
        self.hev_enabled = False

        # FIXME: This looks hacky. Maybe find a better way?
        self.config['n_tpc_pmts'] = self.config['n_he_pmts']

        self.hit_thresholds = self.hit_min_amplitude_he

    def compute(self, raw_records_he, start, end):
        result = super().compute(raw_records_he, start, end)
        return dict(records_he=result['records'],
                    pulse_counts_he=result['pulse_counts'])



import os.path
from unittest import TestCase, skipIf
import numpy as np
import strax
import straxen
import shutil
from straxen.test_utils import nt_test_run_id as test_run_id_nT


class DummyAqmonHits(strax.Plugin):
    """
    Dummy plugin to make some aqmon hits that may span ON and OFF
    signals over chunks

    There are two channels, that signify ON or OFF of a logic signal

    We are interested in the deadtime, which is the time difference
    between ON and OFF singals. This plugin computes that deadtime, even
    if we have missed the first ON (so we start with OFF) or missed the
    last OFF (such that the last signal is an ON). If we miss one ON at
    the start or one OFF at the end, we should just consider all time
    before the first OFF or all the time after the last ON as deadtime.
    """
    vetos_per_chunk = strax.Config(
        default=list(range(1, 10)),
        help='The number of ON/OFF signals per chunk, preferably a '
             'combination of odd and even such that we have both '
             'unmatched ON/OFF singals'
    )
    start_with_channel_on = strax.Config(
        default=True,
        help='If True, start with an ON signal, otherwise start if OFF'
    )
    channel_on = strax.Config(
        default=straxen.AqmonChannels.BUSY_START,
        help='ON channel. Just some channel known to the VetoIntervals plugin.',
        type=int)
    channel_off = strax.Config(
        default=straxen.AqmonChannels.BUSY_STOP,
        help='OFF channel. Just some channel known to the VetoIntervals plugin',
        type=int)
    veto_duration_max = strax.Config(
        default=300e9,
        help='Max duration of one veto [ns].'
    )
    # Start from scratch
    depends_on = ()
    parallel = False
    provides = straxen.AqmonHits.provides
    dtype = straxen.AqmonHits.dtype
    save_when = strax.SaveWhen.NEVER

    # Keep track for we need this from plugin to plugin
    _last_channel_was_off = True
    _last_endtime = 0

    # Will overwrite this with a mutable default in the test
    TOTAL_DEADTIME = None
    TOTAL_SIGNALS = None

    # This value should be larger than the duration of a single hit
    gap_ns_between_chunks = 10

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < len(self.vetos_per_chunk)

    def compute(self, chunk_i):
        if chunk_i == 0:
            self._last_channel_was_off = self.start_with_channel_on

        n_vetos = self.vetos_per_chunk[chunk_i]
        res = np.zeros(n_vetos, self.dtype)

        # Add some randomly increasing times (larger than previous chunk
        res['time'] = self._last_endtime + 1 + np.cumsum(
            np.random.randint(low=2, high=self.veto_duration_max, size=n_vetos)
        )
        res['dt'] = 1
        res['length'] = 1

        if self._last_channel_was_off:
            # Previous chunk, we ended with an off, so now we start with an ON
            res['channel'][::2] = self.channel_on
            res['channel'][1::2] = self.channel_off
            starts = res[::2]['time']
            stops = res[1::2]['time']

            # Sum the time differences between starts and stops to get the deadtime
            self.TOTAL_DEADTIME += [np.sum(stops - starts[:len(stops)])]

        else:
            # Previous chunk, we ended with an ON, so now we start with an OFF
            res['channel'][1::2] = self.channel_on
            res['channel'][::2] = self.channel_off
            starts = res[1::2]['time']
            stops = res[::2]['time']

            # Ignore the first stop (stops last from previous chunk)
            self.TOTAL_DEADTIME += [np.sum(stops[1:] - starts[:len(stops) - 1])]

            # Additionally, add the deadtime that spans the chunks (i.e. the first stop)
            # The gap_ns_between_chunks may look a bit magic, but it's correct
            if chunk_i != 0:
                self.TOTAL_DEADTIME += [self.gap_ns_between_chunks]
            self.TOTAL_DEADTIME += [stops[0] - self._last_endtime]

        # Track the total number of signals both ON and OFF
        self.TOTAL_SIGNALS += [n_vetos]

        previous_end = self._last_endtime
        self._last_endtime = res['time'][-1] + self.gap_ns_between_chunks
        self._last_channel_was_off = res['channel'][-1] == self.channel_off
        if len(res) > 1:
            assert np.sum(res['channel'])

        if chunk_i == len(self.vetos_per_chunk) - 1 and not self._last_channel_was_off:
            # There is one ON without an OFF at the end of the run in
            # the last chunk. This should fill all the way to the end as
            # deadtime:
            dt = self._last_endtime - res['time'][-1]
            self.TOTAL_DEADTIME += [dt]
        return self.chunk(start=previous_end,
                          end=self._last_endtime,
                          data=res)


class DummyEventBasics(strax.Plugin):
    """Get evenly spaced random duration events that don't overlap"""
    n_chunks = strax.Config(default=3)
    event_time_range = strax.Config(default=(0, 3e6),
                                    help='Where to span the durations of chunks over'
                                    )
    events_per_chunk = strax.Config(default=20)
    event_durations = strax.Config(
        default=(1000, 20_000),
        help='event durations (min, max) ns'
    )
    depends_on = ()
    save_when = strax.SaveWhen.ALWAYS
    provides = 'event_basics'
    data_kind = 'events'
    dtype = strax.time_fields
    _events = None

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < self.n_chunks

    def get_events_this_chunk(self, chunk_i):
        if self._events is None:
            res = np.zeros(int(self.events_per_chunk * self.n_chunks), dtype=self.dtype)
            times = np.linspace(*self.event_time_range, len(res))
            res['time'] = times

            endtimes = res['time'] + np.random.randint(*self.event_durations, size=len(res))
            # Don't allow overlapping events
            endtimes[:-1] = np.clip(endtimes[:-1], 0, res['time'][1:] - 1)
            res['endtime'] = endtimes
            self._events = np.split(res, self.n_chunks)
        return self._events[chunk_i]

    def chunk_start(self, chunk_i) -> int:
        # Get the start time of the requested chunk
        assert self._events is not None, "run get_events_this_chunk first!"
        if chunk_i == 0:
            return self._events[chunk_i]['time'][0]
        return self._events[chunk_i - 1]['endtime'][-1]

    def get_chunk_end(self, chunk_i):
        assert self._events is not None, "run get_events_this_chunk first!"
        if chunk_i == self.n_chunks:
            return self.event_time_range[1]
        return self._events[chunk_i]['endtime'][-1]

    def compute(self, chunk_i):
        events = self.get_events_this_chunk(chunk_i)
        events_within = events['endtime'] < self.event_time_range[1]
        events = events[events_within]

        return self.chunk(start=self.chunk_start(chunk_i),
                          end=self.get_chunk_end(chunk_i),
                          data=events)


class TestAqmonProcessing(TestCase):
    def setUp(self) -> None:
        st = straxen.test_utils.nt_test_context().new_context()
        # I'm going to deregister all plugins, since I don't want to
        # get a thousand warnings that some config is not used, make
        # sure to mark all configs as "free options".
        st.set_context_config({'free_options': list(st.config.keys())})
        st._plugin_class_registry = {}

        st.set_config(dict(veto_proximity_window=10 ** 99))
        self.TOTAL_DEADTIME = []
        self.TOTAL_SIGNALS = []

        class DeadTimedDummyAqHits(DummyAqmonHits):
            # Add mutible defaults to give results in the tests below
            TOTAL_DEADTIME = self.TOTAL_DEADTIME
            TOTAL_SIGNALS = self.TOTAL_SIGNALS

        class DummyVi(straxen.VetoIntervals):
            save_when = strax.SaveWhen.ALWAYS

        class DummyVp(straxen.VetoProximity):
            save_when = strax.SaveWhen.NEVER

        st.register(DeadTimedDummyAqHits)
        st.register(DummyVi)
        st.register(DummyVp)
        st.register(DummyEventBasics)
        self.st = st
        self.run = test_run_id_nT
        self.assertFalse(np.sum(self.TOTAL_DEADTIME))
        self.assertFalse(st.is_stored(self.run, 'aqmon_hits'))
        self.assertFalse(st.is_stored(self.run, 'veto_intervals'))

    def test_dummy_plugin_works(self):
        """The simplest plugin """
        self.st.make(self.run, 'aqmon_hits')
        self.assertGreater(np.sum(self.TOTAL_DEADTIME), 0)
        self.assertGreater(np.sum(self.TOTAL_SIGNALS), 0)
        events = self.st.get_array(self.run, 'event_basics')
        self.assertTrue(len(events))

    def test_veto_intervals(self, options=None):
        if options is not None:
            self.st.set_config(options)
        veto_intervals = self.st.get_array(self.run, 'veto_intervals')
        # We should have roughly 1/2 the number of ON/OFF signals of intervals
        # Roughly, since we might have patched an ON/OFF
        self.assertAlmostEqual(len(veto_intervals), sum(self.TOTAL_SIGNALS) / 2, delta=2)
        self.assertTrue(np.sum(self.TOTAL_DEADTIME))
        self.assertEqual(np.sum(veto_intervals['veto_interval']), np.sum(self.TOTAL_DEADTIME))

    def test_veto_intervals_with_missing_on(self):
        self.test_veto_intervals(dict(start_with_channel_on=False))

    def test_make_veto_proximity(self):
        """
        I'm not going to do something fancy here, just checking if we can run the code
        """
        veto_intervals = self.st.get_array(self.run, 'veto_intervals')
        self.st.set_config(dict(event_time_range=[0,
                                                  int(veto_intervals['endtime'][-1])]))
        self.st.make(self.run, 'event_basics')
        for c in self.st.get_iter(self.run, 'veto_proximity'):
            print(c)
        return self.st

    @skipIf(not straxen.test_utils.is_installed('cutax'), 'cutax not installed')
    def test_cut_daq_reader(self):
        st = self.test_make_veto_proximity()
        import cutax
        st.register(cutax.cuts.DAQVeto)
        st.make(self.run, 'cut_daq_veto')

    def tearDown(self) -> None:
        for sf in self.st.storage:
            if not sf.readonly:
                p = getattr(sf, 'path')
                if os.path.exists(p):
                    shutil.rmtree(p)



"""Test for 1T plugins, nT plugins are tested in the ./plugins directory"""
import tempfile
import strax
import straxen
from straxen.test_utils import DummyRawRecords
from immutabledict import immutabledict

test_run_id_1T = '180423_1021'

testing_config_1T = dict(
    hev_gain_model='legacy-to-pe://1T_to_pe_placeholder',
    gain_model='legacy-to-pe://1T_to_pe_placeholder',
    elife=1e6,
    electron_drift_velocity=1e-4,
    electron_drift_time_gate=1700,
)


def _run_plugins(st, make_all=False, run_id=test_run_id_1T, **process_kwargs):
    """
    Try all plugins (except the DAQReader) for a given context (st) to see if
    we can really push some (empty) data from it and don't have any nasty
    problems like that we are referring to some non existant dali folder.
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        st.storage = [strax.DataDirectory(temp_dir)]
        # As we use a temporary directory we should have a clean start
        assert not st.is_stored(run_id, 'raw_records'), 'have RR???'

        if not make_all:
            return

        end_targets = set(st._get_end_targets(st._plugin_class_registry))
        if st.context_config['allow_multiprocess']:
            st.make(run_id, list(end_targets), allow_multiple=True, **process_kwargs)
        else:
            for data_type in end_targets:
                st.make(run_id, data_type)
        # Now make sure we can get some data for all plugins
        all_datatypes = set(st._plugin_class_registry.keys())
        for data_type in all_datatypes:
            savewhen = st._plugin_class_registry[data_type].save_when
            if isinstance(savewhen, (dict, immutabledict)):
                savewhen = savewhen[data_type]
            should_be_stored = savewhen == strax.SaveWhen.ALWAYS
            if should_be_stored:
                is_stored = st.is_stored(run_id, data_type)
                assert is_stored, f"{data_type} did not save correctly!"
    print("Wonderful all plugins work (= at least they don't fail), bye bye")


def _update_context(st, max_workers):
    # Ignore strax-internal warnings
    st.set_context_config({'free_options': tuple(st.config.keys()),
                           'forbid_creation_of': ()})

    st.register(DummyRawRecords)
    st.set_config(testing_config_1T)

    if max_workers - 1:
        st.set_context_config({
            'allow_multiprocess': True,
            'allow_lazy': False,
            'timeout': 120,  # we don't want to build travis for ever
            'allow_shm': strax.processor.SHMExecutor is not None,
        })


def _test_child_options(st, run_id):
    """
    Test which checks if child options are handled correctly.
    """
    # Register all used plugins
    plugins = []
    already_seen = []
    for data_type in st._plugin_class_registry.keys():
        if data_type in already_seen or data_type in straxen.DAQReader.provides:
            continue

        p = st.get_single_plugin(run_id, data_type)
        plugins.append(p)
        already_seen += p.provides

    # Loop over all plugins and check if child options were propagated to the parent:
    for p in plugins:
        for option_name, option in p.takes_config.items():
            # Check if option is a child option:
            if option.child_option:
                # Get corresponding parent option. Do not have to test if
                # parent option name is defined this is already done in strax
                parent_name = option.parent_option_name

                # Now check if parent config was replaced with child:
                t = p.config[parent_name] == p.config[option_name]
                assert t, (f'This is strange the child option "{option_name}" was set to '
                           f'{p.config[option_name]}, but the corresponding parent config'
                           f' "{parent_name}" has the value {p.config[parent_name]}. '
                           f'Please check the options of {p.__class__.__name__} and if '
                           'it is a child plugin (child_plugin=True)!')

                # Test if parent names were removed from the lineage:
                t = parent_name in p.lineage[p.provides[-1]][2]
                assert not t, (f'Found "{parent_name}" in the lineage of {p.__class__.__name__}. '
                               f'This should not have happend since "{parent_name}" is a child of '
                               f'"{option_name}"!')


def test_1T(ncores=1):
    st = straxen.contexts.xenon1t_dali()
    _update_context(st, ncores)
    st.register_all(straxen.legacy.plugins_1t.x1t_cuts)
    for _plugin, _plugin_class in st._plugin_class_registry.items():
        if 'cut' in str(_plugin).lower():
            _plugin_class.save_when = strax.SaveWhen.ALWAYS

    # Run the test
    _run_plugins(st, make_all=True, max_workers=ncores, run_id=test_run_id_1T)
    # set all the configs to be non-CMT
    st.set_config(testing_config_1T)
    _test_child_options(st, test_run_id_1T)

    print(st.context_config)

def test_1T_mc():
    # Run multicore
    test_1T(2)



import socket
import unittest
import strax
import os
import straxen
import shutil
import json
from bson import json_util


class TestRucioLocal(unittest.TestCase):
    """
    Test the behavoir of how the Rucio Local frontend should behave
    """

    def setUp(self) -> None:
        self.test_keys = [
            strax.DataKey(run_id=run_id,
                          data_type='dtype',
                          lineage={'dtype': ['Plugin', '0.0.0', {}], }
                          )
            for run_id in ('-1', '-2')
        ]
        self.rucio_path = './.test_rucio'
        self.write_test_data()

    def tearDown(self) -> None:
        shutil.rmtree(self.rucio_path)

    def test_find(self):
        rucio_local = straxen.RucioLocalFrontend(path=self.rucio_path)
        find_result = rucio_local.find(self.test_keys[0])
        assert len(find_result) and find_result[0] == 'RucioLocalBackend', find_result

    def test_find_several(self):
        rucio_local = straxen.RucioLocalFrontend(path=self.rucio_path)
        find_several_results = rucio_local.find_several(self.test_keys)
        assert find_several_results, find_several_results
        for find_result in find_several_results:
            assert len(find_result) and find_result[0] == 'RucioLocalBackend', find_result

    def test_find_fuzzy(self):
        changed_keys = []
        rucio_local = straxen.RucioLocalFrontend(path=self.rucio_path)
        for key in self.test_keys:
            changed_key = strax.DataKey(
                run_id=key.run_id,
                data_type=key.data_type,
                lineage={'dtype': ['Plugin', '1.0.0', {}], }
            )
            changed_keys += [changed_key]

            # We shouldn't find this data
            with self.assertRaises(strax.DataNotAvailable):
                rucio_local.find(changed_key)

        # Also find several shouldn't work
        find_several_keys = rucio_local.find_several(changed_keys)
        self.assertFalse(any(find_several_keys))

        # Now test fuzzy
        with self.assertWarns(UserWarning):
            find_several_keys_fuzzy = rucio_local.find_several(
                changed_keys,
                fuzzy_for=changed_keys[0].data_type,
            )
        self.assertTrue(all(find_several_keys_fuzzy))

    def write_test_data(self):
        os.makedirs(self.rucio_path, exist_ok=True)
        for key in self.test_keys:
            did = straxen.key_to_rucio_did(key)
            metadata = {
                'writing_ended': 1,
                'chunks': [{"filename": f"{key.data_type}-{key.lineage_hash}-000000",
                            },
                           ],
                "lineage_hash": key.lineage_hash,
                "lineage": key.lineage,
            }
            self.write_md(self.rucio_path, did, metadata)
            self.write_chunks(self.rucio_path, did, [c['filename'] for c in metadata['chunks']])

    @staticmethod
    def write_md(rucio_path, did, content: dict):
        md_did = f'{did}-metadata.json'
        md_path = straxen.storage.rucio_local.rucio_path(rucio_path, md_did)
        os.makedirs(os.path.split(md_path)[0], exist_ok=True)
        with open(md_path, mode='w') as f:
            f.write(json.dumps(content, default=json_util.default))

    @staticmethod
    def write_chunks(rucio_path, did, file_names):
        for file_name in file_names:
            file_did = did.split(':')[0] + ":" + file_name
            chunk_path = straxen.storage.rucio_local.rucio_path(rucio_path, file_did)
            os.makedirs(os.path.split(chunk_path)[0], exist_ok=True)
            with open(chunk_path, mode='w') as f:
                f.write(file_name)


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
class TestBasics(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        For testing purposes, slightly alter the RucioFrontend such that
         we can run tests outside of dali too
        """
        # Some non-existing keys that we will try finding in the test cases.
        cls.test_keys = [
            strax.DataKey(run_id=run_id,
                          data_type='dtype',
                          lineage={'dtype': ['Plugin', '0.0.0.', {}], }
                          )
            for run_id in ('-1', '-2')
        ]

    def test_load_context_defaults(self):
        """Don't fail immediately if we start a context due to Rucio"""
        st = straxen.contexts.xenonnt_online(minimum_run_number=10_000,
                                             maximum_run_number=10_010,
                                             )
        st.select_runs()

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_find_several_remote(self):
        """
        Let's try running a find_several with the include remote.
        This should fail but when no rucio is installed or else it
        shouldn't find any data.
        """
        try:
            rucio = straxen.RucioRemoteFrontend()
        except ImportError:
            pass
        else:
            found = rucio.find_several(self.test_keys)
            # We shouldn't find any of these
            assert found == [False for _ in self.test_keys]

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_find_local_rundb(self):
        """Make sure that we don't find the non existing data"""
        run_db = straxen.RunDB(rucio_path='./rucio_test')
        with self.assertRaises(strax.DataNotAvailable):
            run_db.find(self.test_keys[0])

    def test_determine_rse(self):
        class DummyLocalRucio(straxen.RucioLocalFrontend):
            local_prefixes = {'any': './any'}
            local_rses = {}

        dummy_class = DummyLocalRucio(path='./')
        assert dummy_class.determine_rse() is None

        update = {'any': socket.getfqdn()}
        dummy_class.local_rses = update
        assert dummy_class.determine_rse() == 'any'

        # now the init should also work
        DummyLocalRucio.local_rses = update
        assert DummyLocalRucio().path == './any'

        with self.assertRaises(ValueError):
            dummy_class.local_rses.update({'some other!': socket.getfqdn()})
            dummy_class.determine_rse()

    @unittest.skipIf(not straxen.utilix_is_configured(), "No DB access")
    @unittest.skipIf(socket.getfqdn() in straxen.RucioLocalFrontend.local_rses,
                     "Testing useless frontends only works on hosts where it's not supposed to work")
    def test_useless_frontend(self):
        """
        Test that using a rucio-local frontend on a non-RSE listed site
        doesn't cause issues when registered
        """
        rucio_local = straxen.RucioLocalFrontend()
        assert rucio_local.path is None
        with self.assertRaises(strax.DataNotAvailable):
            rucio_local.find(self.test_keys[0])
        # Do a small test that we did not break everything by having a useless fontend
        st = straxen.test_utils.nt_test_context(minimum_run_number=10_000,
                                                maximum_run_number=10_005,
                                                include_rucio_local=True,
                                                keep_default_storage=True,
                                                )
        self.assertTrue(
            str(rucio_local.__class__) in [str(sf.__class__) for sf in st.storage],
            "Rucio local did not get registered??"
        )
        st.select_runs()



from . import peak_top_bottom_params
from .peak_top_bottom_params import *

from . import peak_ambience
from .peak_ambience import *

from . import peak_basics
from .peak_basics import *

from . import peak_classification_bayes
from .peak_classification_bayes import *

from . import peak_s1_positions_cnn
from .peak_s1_positions_cnn import *

from . import peak_positions
from .peak_positions import *

from . import peak_positions_cnn
from .peak_positions_cnn import *

from . import peak_positions_gcn
from .peak_positions_gcn import *

from . import peak_positions_mlp
from .peak_positions_mlp import *

from . import peak_proximity
from .peak_proximity import *

from . import peak_shadow
from .peak_shadow import *

from . import peaks
from .peaks import *

from . import peak_per_event
from .peak_per_event import *

from . import peak_corrections
from .peak_corrections import *

from . import peaks_subtyping
from .peaks_subtyping import *



from hypothesis import strategies, given, settings, example
from straxen.plugins.events.event_pattern_fit import binom_test, lbinom_pmf_mode, lbinom_pmf, lbinom_pmf_inverse, binom_sf
import numpy as np
import scipy.stats as sps

@settings(deadline=None)
@given(strategies.floats(0., 1.), strategies.floats(0.01, 0.99), strategies.floats(2., 1000))
def test_patternfit_stats(aftobs, aft, s1tot):
    s1top = aftobs*s1tot
    assert( (binom_test(s1top, s1tot, aft) >= 0) & (binom_test(s1top, s1tot, aft) <= 1) )

@settings(deadline=None)
@given(strategies.floats(0., 1.), strategies.floats(0.01, 0.99), strategies.floats(2., 1000))
def test_inverse_pdf(aftobs, aft, s1tot):
    s1top = aftobs * s1tot
    #code stolen from binom_test, checking if the other point on

    k = s1top
    n = s1tot
    p = aft

    mode = lbinom_pmf_mode(0, n, 0, (n, p))
    distance = abs(mode - k)
    target = lbinom_pmf(k, n, p)

    if k < mode: 
        j_min = mode
        j_max = min(mode + 2.0 * distance, n)
        j = lbinom_pmf_inverse(j_min, j_max, target, (n, p))
    else:
        j_min = max(mode - 2.0 * distance, 0)
        j_max = mode
        j = lbinom_pmf_inverse(j_min, j_max, target, (n, p))  
    if np.isnan(j) or j == n:
        #if no other crossing is found, we cannot check if the two give same pdf
        assert(True)
    else:
        pdf_value = lbinom_pmf(j, n, p)
        #values in log, it is not _really_ crucial that they're precisely equal
        np.testing.assert_almost_equal(pdf_value, target, decimal=2) 

@settings(deadline=None)
@strategies.composite 
def n_and_k(draw):
    n = draw(strategies.integers(2, 1000))
    k = draw(strategies.integers(0, n))
    return n,k

@settings(deadline=None)
@given(nk = n_and_k(), p = strategies.floats(0.01, 0.99))
@example((10, 5), 0.5)
def test_pmf(nk, p):
    #test that at integer n, k binom_pmf agrees with the scipy result
    n, k = nk
    sps_lpmf = sps.binom(n, p).logpmf(k)
    straxen_lpmf = lbinom_pmf(k, n, p)
    np.testing.assert_almost_equal(straxen_lpmf, sps_lpmf, decimal=2)

@settings(deadline=None)
@given(nk = n_and_k(), p = strategies.floats(0.01, 0.99))
def test_pvalue(nk, p):
    #test that at integer n, k binom_pmf agrees with the scipy result
    n, k = nk
    sps_pval = sps.binom(n, p).sf(k + 0.1)
    straxen_pval = binom_sf(k, n, p)
    np.testing.assert_almost_equal(straxen_pval, sps_pval, decimal=2)



"""Position reconstruction for Xenon-nT"""

import numpy as np
import strax
import straxen
from warnings import warn

export, __all__ = strax.exporter()

DEFAULT_POSREC_ALGO = "mlp"


@export
class PeakPositionsBaseNT(strax.Plugin):
    """
    Base class for reconstructions.
    This class should only be used when subclassed for the different
    algorithms. Provides x_algorithm, y_algorithm for all peaks > than
    min-reconstruction area based on the top array.
    """
    __version__ = '0.0.0'

    depends_on = ('peaks',)
    algorithm = None
    compressor = 'zstd'
    parallel = True  # can set to "process" after #82

    min_reconstruction_area = straxen.URLConfig(
        help='Skip reconstruction if area (PE) is less than this',
        default=10, infer_type=False, )

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, infer_type=False,
                                   help="Number of top PMTs")

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')
        dtype = [('x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 X position (cm), uncorrected'),
                 ('y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 Y position (cm), uncorrected')]
        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """
        Simple wrapper to have several tf_model_mlp, tf_model_cnn, ..
        point to this same function in the compute method
        """
        model = getattr(self, f'tf_model_{self.algorithm}', None)
        if model is None:
            warn(f'Setting model to None for {self.__class__.__name__} will '
                 f'set only nans as output for {self.algorithm}')
        if isinstance(model, str):
            raise ValueError(f'open files from tf:// protocol! Got {model} '
                             f'instead, see tests/test_posrec.py for examples.')
        return model

    def compute(self, peaks):
        result = np.ones(len(peaks), dtype=self.dtype)
        result['time'], result['endtime'] = peaks['time'], strax.endtime(peaks)

        result['x_' + self.algorithm] *= float('nan')
        result['y_' + self.algorithm] *= float('nan')
        model = self.get_tf_model()

        if model is None:
            # This plugin is disabled since no model is provided
            return result

        # Keep large peaks only
        peak_mask = peaks['area'] > self.min_reconstruction_area
        if not np.sum(peak_mask):
            # Nothing to do, and .predict crashes on empty arrays
            return result

        # Getting actual position reconstruction
        area_per_channel_top = peaks['area_per_channel'][
                               peak_mask,
                               0:self.n_top_pmts]
        with np.errstate(divide='ignore', invalid='ignore'):
            area_per_channel_top = (
                    area_per_channel_top /
                    np.max(area_per_channel_top, axis=1).reshape(-1, 1)
            )
        area_per_channel_top = area_per_channel_top.reshape(-1,
                                                            self.n_top_pmts
                                                            )
        output = model.predict(area_per_channel_top, verbose=0)

        # writing output to the result
        result['x_' + self.algorithm][peak_mask] = output[:, 0]
        result['y_' + self.algorithm][peak_mask] = output[:, 1]
        return result



import strax
import straxen
import numpy as np
from warnings import warn
export, __all__ = strax.exporter()


@export
class EventS2PositionBase(strax.Plugin):
    """
    Base pluging for S2 position reconstruction at event level
    """
    __version__ = '0.0.0'
    depends_on = ('event_area_per_channel', 'event_basics')

    algorithm = None
    compressor = 'zstd'
    parallel = True  # can set to "process" after #82

    min_reconstruction_area = straxen.URLConfig(
        help='Skip reconstruction if area (PE) is less than this',
        default=10, infer_type=False, )
    n_top_pmts = straxen.URLConfig(
        default=straxen.n_top_pmts, infer_type=False,
        help="Number of top PMTs")

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')

        dtype = [('event_s2_x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 X position (cm), uncorrected'),
                 ('event_s2_y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 Y position (cm), uncorrected'),
                 ('event_alt_s2_x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} alt S2 X position (cm), uncorrected'),
                 ('event_alt_s2_y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} alt S2 Y position (cm), uncorrected')]
        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """
        Simple wrapper to have several tf_event_model_mlp, tf_event_model_cnn, ..
        point to this same function in the compute method
        """
        model = getattr(self, f'tf_event_model_{self.algorithm}', None)
        if model is None:
            warn(f'Setting model to None for {self.__class__.__name__} will '
                 f'set only nans as output for {self.algorithm}')
        if isinstance(model, str):
            raise ValueError(f'open files from tf:// protocol! Got {model} '
                             f'instead, see tests/test_posrec.py for examples.')
        return model

    def compute(self, events):
        result = np.ones(len(events), dtype=self.dtype)
        result['time'], result['endtime'] = events['time'], strax.endtime(events)
        result['event_s2_x_' + self.algorithm] *= float('nan')
        result['event_s2_y_' + self.algorithm] *= float('nan')
        result['event_alt_s2_x_' + self.algorithm] *= float('nan')
        result['event_alt_s2_y_' + self.algorithm] *= float('nan')

        model = self.get_tf_model()

        for p_type in ['s2', 'alt_s2']:
            peak_mask = events[p_type+'_area'] > self.min_reconstruction_area
            if not np.sum(peak_mask):
                continue

            _top_pattern = events[p_type + '_area_per_channel'][peak_mask, 0:self.n_top_pmts]
            with np.errstate(divide='ignore', invalid='ignore'):
                _top_pattern = (
                        _top_pattern /
                        np.max(_top_pattern, axis=1).reshape(-1, 1)
                )

            _top_pattern = _top_pattern.reshape(-1,self.n_top_pmts)
            _pos=model.predict(_top_pattern)

            result['event_'+p_type+'_x_' + self.algorithm][peak_mask] = _pos[:, 0]
            result['event_'+p_type+'_y_' + self.algorithm][peak_mask] = _pos[:, 1]

        return result



import strax

from straxen.plugins.peaklets.peaklet_classification import PeakletClassification
from straxen.plugins.defaults import HE_PREAMBLE

export, __all__ = strax.exporter()


@export
class PeakletClassificationHighEnergy(PeakletClassification):
    __doc__ = HE_PREAMBLE + PeakletClassification.__doc__
    provides = 'peaklet_classification_he'
    depends_on = ('peaklets_he',)
    __version__ = '0.0.2'
    child_plugin = True

    def compute(self, peaklets_he):
        return super().compute(peaklets_he)



import strax
import straxen
from straxen import pax_file, first_sr1_run
from straxen.common import get_resource
import tempfile
import numpy as np
import json
import os

export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option(
        'nn_architecture', infer_type=False,
        help='Path to JSON of neural net architecture',
        default_by_run=[
            (0, pax_file('XENON1T_tensorflow_nn_pos_20171217_sr0.json')),
            (first_sr1_run, straxen.aux_repo + '3548132b55f81a43654dba5141366041e1daaf01/strax_files/XENON1T_tensorflow_nn_pos_20171217_sr1_reformatted.json')]),  # noqa
    strax.Option(
        'nn_weights', infer_type=False,
        help='Path to HDF5 of neural net weights',
        default_by_run=[
            (0, pax_file('XENON1T_tensorflow_nn_pos_weights_20171217_sr0.h5')),
            (first_sr1_run, pax_file('XENON1T_tensorflow_nn_pos_weights_20171217_sr1.h5'))]),  # noqa
    strax.Option('min_reconstruction_area',
                 help='Skip reconstruction if area (PE) is less than this',
                 default=10, infer_type=False, ),
    strax.Option('n_top_pmts', default=straxen.n_top_pmts, infer_type=False,
                 help="Number of top PMTs")
)
class PeakPositions1T(strax.Plugin):
    """Compute the S2 (x,y)-position based on a neural net."""
    dtype = [('x', np.float32,
              'Reconstructed S2 X position (cm), uncorrected'),
             ('y', np.float32,
              'Reconstructed S2 Y position (cm), uncorrected')
             ] + strax.time_fields
    depends_on = ('peaks',)
    provides = "peak_positions"

    # Parallelization doesn't seem to make it go faster
    # Is there much pure-python stuff in tensorflow?
    # Process-level paralellization might work, but you'd have to do setup
    # in each process, which probably negates the benefits,
    # except for huge chunks
    parallel = False

    __version__ = '0.1.1'

    def setup(self):
        import tensorflow as tf
        keras = tf.keras
        nn_conf = get_resource(self.config['nn_architecture'], fmt='json')
        # badPMTList was inserted by a very clever person into the keras json
        # file. Let's delete it to prevent future keras versions from crashing.
        # Do NOT try `del nn_conf['badPMTList']`! See get_resource docstring
        # for the gruesome details.
        bad_pmts = nn_conf['badPMTList']
        nn = keras.models.model_from_json(json.dumps({
            k: v
            for k, v in nn_conf.items()
            if k != 'badPMTList'}))
        self.pmt_mask = ~np.in1d(np.arange(self.config['n_top_pmts']),
                                 bad_pmts)

        # Keras needs a file to load its weights. We can't put the load
        # inside the context, then it would break on Windows,
        # because there temporary files cannot be opened again.
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(get_resource(self.config['nn_weights'],
                                 fmt='binary'))
            fname = f.name
        nn.load_weights(fname)
        os.remove(fname)
        self.nn = nn

    def compute(self, peaks):
        result = np.ones(len(peaks), dtype=self.dtype)
        result['time'], result['endtime'] = peaks['time'], strax.endtime(peaks)
        result['x'] *= float('nan')
        result['y'] *= float('nan')

        # Keep large peaks only
        peak_mask = peaks['area'] > self.config['min_reconstruction_area']
        if not np.sum(peak_mask):
            # Nothing to do, and .predict crashes on empty arrays
            return result

        # Input: normalized hitpatterns in good top PMTs
        _in = peaks['area_per_channel'][peak_mask, :]
        _in = _in[:, :self.config['n_top_pmts']][:, self.pmt_mask]
        with np.errstate(divide='ignore', invalid='ignore'):
            _in /= _in.sum(axis=1).reshape(-1, 1)

        # Output: positions in mm (unfortunately), so convert to cm
        _out = self.nn.predict(_in) / 10

        # Set output in valid rows. Do NOT try result[peak_mask]['x']
        # unless you want all NaN positions (boolean masks make a copy unless
        # they are used as the last index)
        result['x'][peak_mask] = _out[:, 0]
        result['y'][peak_mask] = _out[:, 1]
        return result



import numba
import numpy as np
import strax
from immutabledict import immutabledict
from strax.processing.general import _touching_windows
from strax.dtypes import DIGITAL_SUM_WAVEFORM_CHANNEL
import straxen


export, __all__ = strax.exporter()


@export
class Peaklets(strax.Plugin):
    """
    Split records into:
     - peaklets
     - lone_hits

    Peaklets are very aggressively split peaks such that we are able
    to find S1-S2s even if they are close to each other. (S2) Peaks
    that are split into too many peaklets will be merged later on.

    To get Peaklets from records apply/do:
        1. Hit finding
        2. Peak finding
        3. Peak splitting using the natural breaks algorithm
        4. Compute the digital sum waveform

    Lone hits are all hits which are outside of any peak. The area of
    lone_hits includes the left and right hit extension, except the
    extension overlaps with any peaks or other hits.
    """
    depends_on = ('records',)
    provides = ('peaklets', 'lone_hits')
    data_kind = dict(peaklets='peaklets',
                     lone_hits='lone_hits')
    parallel = 'process'
    compressor = 'zstd'

    __version__ = '1.1.0'

    peaklet_gap_threshold = straxen.URLConfig(
        default=700, infer_type=False,
        help="No hits for this many ns triggers a new peak")

    peak_left_extension = straxen.URLConfig(
        default=30, infer_type=False,
        help="Include this many ns left of hits in peaks")

    peak_right_extension = straxen.URLConfig(
        default=200, infer_type=False,
        help="Include this many ns right of hits in peaks")

    peak_min_pmts = straxen.URLConfig(
        default=2, infer_type=False,
        help="Minimum number of contributing PMTs needed to define a peak")

    peak_split_gof_threshold = straxen.URLConfig(
        # See https://xe1t-wiki.lngs.infn.it/doku.php?id=
        # xenon:xenonnt:analysis:strax_clustering_classification
        # #natural_breaks_splitting
        # for more information
        default=(
            None,  # Reserved
            ((0.5, 1.0), (6.0, 0.4)),
            ((2.5, 1.0), (5.625, 0.4))), infer_type=False,
        help='Natural breaks goodness of fit/split threshold to split '
             'a peak. Specify as tuples of (log10(area), threshold).')

    peak_split_filter_wing_width = straxen.URLConfig(
        default=70, infer_type=False,
        help='Wing width of moving average filter for '
             'low-split natural breaks')

    peak_split_min_area = straxen.URLConfig(
        default=40., infer_type=False,
        help='Minimum area to evaluate natural breaks criterion. '
             'Smaller peaks are not split.')

    peak_split_iterations = straxen.URLConfig(
        default=20, infer_type=False,
        help='Maximum number of recursive peak splits to do.')

    diagnose_sorting = straxen.URLConfig(
        track=False, default=False, infer_type=False,
        help="Enable runtime checks for sorting and disjointness")

    gain_model = straxen.URLConfig(
        infer_type=False,
        help='PMT gain model. Specify as URL or explicit value'
    )

    tight_coincidence_window_left = straxen.URLConfig(
        default=50, infer_type=False,
        help="Time range left of peak center to call a hit a tight coincidence (ns)")

    tight_coincidence_window_right = straxen.URLConfig(
        default=50, infer_type=False,
        help="Time range right of peak center to call a hit a tight coincidence (ns)")

    n_tpc_pmts = straxen.URLConfig(
        type=int,
        help='Number of TPC PMTs')

    n_top_pmts = straxen.URLConfig(
        type=int,
        help="Number of top TPC array PMTs")

    sum_waveform_top_array = straxen.URLConfig(
        default=True,
        type=bool,
        help='Digitize the sum waveform of the top array separately'
    )

    saturation_correction_on = straxen.URLConfig(
        default=True, infer_type=False,
        help='On off switch for saturation correction')

    saturation_reference_length = straxen.URLConfig(
        default=100, infer_type=False,
        help="Maximum number of reference sample used "
             "to correct saturated samples")

    saturation_min_reference_length = straxen.URLConfig(
        default=20, infer_type=False,
        help="Minimum number of reference sample used "
             "to correct saturated samples")

    peaklet_max_duration = straxen.URLConfig(
        default=int(10e6), infer_type=False,
        help="Maximum duration [ns] of a peaklet")

    channel_map = straxen.URLConfig(
        track=False, type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) "
             "channel number.")

    hit_min_amplitude = straxen.URLConfig(
        track=True, infer_type=False,
        default='cmt://hit_thresholds_tpc?version=ONLINE&run_id=plugin.run_id',
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_tpc_pmts, or a number,'
             'or a string like "pmt_commissioning_initial" which means calling'
             'hitfinder_thresholds.py'
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.'
    )

    def infer_dtype(self):
        return dict(
            peaklets=strax.peak_dtype(
                n_channels=self.n_tpc_pmts,
                digitize_top=self.sum_waveform_top_array,
            ),
            lone_hits=strax.hit_dtype,
        )

    def setup(self):
        if self.peak_min_pmts > 2:
            # Can fix by re-splitting,
            raise NotImplementedError(
                f"Raising the peak_min_pmts to {self.peak_min_pmts} "
                f"interferes with lone_hit definition. "
                f"See github.com/XENONnT/straxen/issues/295")

        self.to_pe = self.gain_model

        self.hit_thresholds = self.hit_min_amplitude

        self.channel_range = self.channel_map['tpc']

    def compute(self, records, start, end):
        r = records

        hits = strax.find_hits(r, min_amplitude=self.hit_thresholds)

        # Remove hits in zero-gain channels
        # they should not affect the clustering!
        hits = hits[self.to_pe[hits['channel']] != 0]

        hits = strax.sort_by_time(hits)

        # Use peaklet gap threshold for initial clustering
        # based on gaps between hits
        peaklets = strax.find_peaks(
            hits, self.to_pe,
            gap_threshold=self.peaklet_gap_threshold,
            left_extension=self.peak_left_extension,
            right_extension=self.peak_right_extension,
            min_channels=self.peak_min_pmts,
            # NB, need to have the data_top field here, will discard if not digitized later
            result_dtype=strax.peak_dtype(n_channels=self.n_tpc_pmts, digitize_top=True),
            max_duration=self.peaklet_max_duration,
        )

        # Make sure peaklets don't extend out of the chunk boundary
        # This should be very rare in normal data due to the ADC pretrigger
        # window.
        self.clip_peaklet_times(peaklets, start, end)

        # Get hits outside peaklets, and store them separately.
        # fully_contained is OK provided gap_threshold > extension,
        # which is asserted inside strax.find_peaks.
        is_lone_hit = strax.fully_contained_in(hits, peaklets) == -1
        lone_hits = hits[is_lone_hit]
        strax.integrate_lone_hits(
            lone_hits, records, peaklets,
            save_outside_hits=(self.peak_left_extension,
                               self.peak_right_extension),
            n_channels=len(self.to_pe))

        # Compute basic peak properties -- needed before natural breaks
        hits = hits[~is_lone_hit]
        # Define regions outside of peaks such that _find_hit_integration_bounds
        # is not extended beyond a peak.
        outside_peaks = self.create_outside_peaks_region(peaklets, start, end)
        strax.find_hit_integration_bounds(
            hits, outside_peaks, records,
            save_outside_hits=(self.peak_left_extension,
                               self.peak_right_extension),
            n_channels=len(self.to_pe),
            allow_bounds_beyond_records=True,
        )

        # Transform hits to hitlets for naming conventions. A hit refers
        # to the central part above threshold a hitlet to the entire signal
        # including the left and right extension.
        # (We are not going to use the actual hitlet data_type here.)
        hitlets = hits
        del hits

        hitlets['time'] -= (hitlets['left'] - hitlets['left_integration']) * hitlets['dt']
        hitlets['length'] = hitlets['right_integration'] - hitlets['left_integration']
        hitlets = strax.sort_by_time(hitlets)
        rlinks = strax.record_links(records)

        # If sum_waveform_top_array is false, don't digitize the top array
        n_top_pmts_if_digitize_top = self.n_top_pmts if self.sum_waveform_top_array else -1
        strax.sum_waveform(peaklets, hitlets, r, rlinks, self.to_pe, n_top_channels=n_top_pmts_if_digitize_top)

        strax.compute_widths(peaklets)

        # Split peaks using low-split natural breaks;
        # see https://github.com/XENONnT/straxen/pull/45
        # and https://github.com/AxFoundation/strax/pull/225
        peaklets = strax.split_peaks(
            peaklets, hitlets, r, rlinks, self.to_pe,
            algorithm='natural_breaks',
            threshold=self.natural_breaks_threshold,
            split_low=True,
            filter_wing_width=self.peak_split_filter_wing_width,
            min_area=self.peak_split_min_area,
            do_iterations=self.peak_split_iterations,
            n_top_channels=n_top_pmts_if_digitize_top
        )

        # Saturation correction using non-saturated channels
        # similar method used in pax
        # see https://github.com/XENON1T/pax/pull/712
        # Cases when records is not writeable for unclear reason
        # only see this when loading 1T test data
        # more details on https://numpy.org/doc/stable/reference/generated/numpy.ndarray.flags.html
        if not r['data'].flags.writeable:
            r = r.copy()

        if self.saturation_correction_on:
            peak_list = peak_saturation_correction(
                r, rlinks, peaklets, hitlets, self.to_pe,
                reference_length=self.saturation_reference_length,
                min_reference_length=self.saturation_min_reference_length,
                n_top_channels=n_top_pmts_if_digitize_top,
            )

            # Compute the width again for corrected peaks
            strax.compute_widths(peaklets, select_peaks_indices=peak_list)

        hitlet_time_shift = (hitlets['left'] - hitlets['left_integration']) * hitlets['dt']
        hit_max_times = hitlets['time'] + hitlet_time_shift  # add time shift again to get correct maximum
        hit_max_times += hitlets['dt'] * hit_max_sample(records, hitlets)

        # Compute tight coincidence level.
        # Making this a separate plugin would
        # (a) doing hitfinding yet again (or storing hits)
        # (b) increase strax memory usage / max_messages,
        #     possibly due to its currently primitive scheduling.
        hit_max_times_argsort = np.argsort(hit_max_times)
        sorted_hit_max_times = hit_max_times[hit_max_times_argsort]
        sorted_hit_channels = hitlets['channel'][hit_max_times_argsort]
        peaklet_max_times = (
            peaklets['time']
            + np.argmax(peaklets['data'], axis=1) * peaklets['dt'])
        peaklets['tight_coincidence'] = get_tight_coin(
            sorted_hit_max_times,
            sorted_hit_channels,
            peaklet_max_times,
            self.tight_coincidence_window_left,
            self.tight_coincidence_window_right,
            self.channel_range)

        # Add max and min time difference between apexes of hits
        self.add_hit_features(hitlets, hit_max_times, peaklets)

        if self.diagnose_sorting and len(r):
            assert np.diff(r['time']).min(initial=1) >= 0, "Records not sorted"
            assert np.diff(hitlets['time']).min(initial=1) >= 0, "Hits/Hitlets not sorted"
            assert np.all(peaklets['time'][1:]
                          >= strax.endtime(peaklets)[:-1]), "Peaks not disjoint"

        # Update nhits of peaklets:
        counts = strax.touching_windows(hitlets, peaklets)
        counts = np.diff(counts, axis=1).flatten()
        peaklets['n_hits'] = counts

        # Drop the data_top field
        if n_top_pmts_if_digitize_top <= 0:
            peaklets = drop_data_top_field(peaklets, self.dtype_for('peaklets'))

        # Check channel of peaklets
        peaklets_unique_channel = np.unique(peaklets['channel'])
        if (peaklets_unique_channel == DIGITAL_SUM_WAVEFORM_CHANNEL).sum() > 1:
            raise ValueError(
                f'Found channel number of peaklets other than {DIGITAL_SUM_WAVEFORM_CHANNEL}')
        # Check tight_coincidence
        if not np.all(peaklets['n_hits'] >= peaklets['tight_coincidence']):
            raise ValueError(
                f'Found n_hits less than tight_coincidence')

        return dict(peaklets=peaklets,
                    lone_hits=lone_hits)

    def natural_breaks_threshold(self, peaks):
        rise_time = -peaks['area_decile_from_midpoint'][:, 1]

        # This is ~1 for an clean S2, ~0 for a clean S1,
        # and transitions gradually in between.
        f_s2 = 8 * np.log10(rise_time.clip(1, 1e5) / 100)
        f_s2 = 1 / (1 + np.exp(-f_s2))

        log_area = np.log10(peaks['area'].clip(1, 1e7))
        thresholds = self.peak_split_gof_threshold
        return (
                f_s2 * np.interp(
            log_area,
            *np.transpose(thresholds[2]))
                + (1 - f_s2) * np.interp(
            log_area,
            *np.transpose(thresholds[1])))

    @staticmethod
    @numba.njit(nogil=True, cache=True)
    def clip_peaklet_times(peaklets, start, end):
        for p in peaklets:
            if p['time'] < start:
                p['time'] = start
            if strax.endtime(p) > end:
                p['length'] = (end - p['time']) // p['dt']

    @staticmethod
    def create_outside_peaks_region(peaklets, start, end):
        """
        Creates time intervals which are outside peaks.

        :param peaklets: Peaklets for which intervals should be computed.
        :param start: Chunk start
        :param end: Chunk end
        :return: array of strax.time_fields dtype.
        """
        if not len(peaklets):
            return np.zeros(0, dtype=strax.time_fields)

        outside_peaks = np.zeros(len(peaklets) + 1,
                                 dtype=strax.time_fields)

        outside_peaks[0]['time'] = start
        outside_peaks[0]['endtime'] = peaklets[0]['time']
        outside_peaks[1:-1]['time'] = strax.endtime(peaklets[:-1])
        outside_peaks[1:-1]['endtime'] = peaklets['time'][1:]
        outside_peaks[-1]['time'] = strax.endtime(peaklets[-1])
        outside_peaks[-1]['endtime'] = end
        return outside_peaks

    @staticmethod
    def add_hit_features(hitlets, hit_max_times, peaklets):
        """
        Create hits timing features
        :param hitlets_max: hitlets with only max height time.
        :param peaklets: Peaklets for which intervals should be computed.
        :return: array of peaklet_timing dtype.
        """
        hits_w_max = np.zeros(
            len(hitlets),
            strax.merged_dtype(
                [np.dtype([('max_time', np.int64)]), np.dtype(strax.time_fields)]))
        hits_w_max['time'] = hitlets['time']
        hits_w_max['endtime'] = strax.endtime(hitlets)
        hits_w_max['max_time'] = hit_max_times
        split_hits = strax.split_by_containment(hits_w_max, peaklets)
        for peaklet, h_max in zip(peaklets, split_hits):
            max_time_diff = np.diff(np.sort(h_max['max_time']))
            if len(max_time_diff) > 0:
                peaklet['max_diff'] = max_time_diff.max()
                peaklet['min_diff'] = max_time_diff.min()
            else:
                peaklet['max_diff'] = -1
                peaklet['min_diff'] = -1

def drop_data_top_field(peaklets, goal_dtype, _name_function= '_drop_data_top_field'):
    """Return peaklets without the data_top field"""
    peaklets_without_top_field = np.zeros(len(peaklets), dtype=goal_dtype)
    strax.copy_to_buffer(peaklets, peaklets_without_top_field, _name_function)
    del peaklets
    return peaklets_without_top_field


@numba.jit(nopython=True, nogil=True, cache=False)
def peak_saturation_correction(records, rlinks, peaks, hitlets, to_pe,
                               reference_length=100,
                               min_reference_length=20,
                               use_classification=False,
                               n_top_channels=0,
                               ):
    """Correct the area and per pmt area of peaks from saturation
    :param records: Records
    :param rlinks: strax.record_links of corresponding records.
    :param peaks: Peaklets / Peaks
    :param hitlets: Hitlets found in records to build peaks.
        (Hitlets are hits including the left/right extension)
    :param to_pe: adc to PE conversion (length should equal number of PMTs)
    :param reference_length: Maximum number of reference sample used
    to correct saturated samples
    :param min_reference_length: Minimum number of reference sample used
    to correct saturated samples
    :param use_classification: Option of using classification to pick only S2
    :param n_top_channels: Number of top array channels.
    """

    if not len(records):
        return
    if not len(peaks):
        return

    # Search for peaks with saturated channels
    mask = peaks['n_saturated_channels'] > 0
    if use_classification:
        mask &= peaks['type'] == 2
    peak_list = np.where(mask)[0]
    # Look up records that touch each peak
    record_ranges = _touching_windows(
        records['time'],
        strax.endtime(records),
        peaks[peak_list]['time'],
        strax.endtime(peaks[peak_list]))

    # Create temporary arrays for calculation
    dt = records[0]['dt']
    n_channels = len(peaks[0]['saturated_channel'])
    len_buffer = np.max(peaks['length'] * peaks['dt']) // dt + 1
    max_nrecord = len_buffer // len(records[0]['data']) + 1

    # Buff the sum wf [pe] of non-saturated channels
    b_sumwf = np.zeros(len_buffer, dtype=np.float32)
    # Buff the records 'data' [ADC] in saturated channels
    b_pulse = np.zeros((n_channels, len_buffer), dtype=np.int16)
    # Buff the corresponding record index of saturated channels
    b_index = np.zeros((n_channels, max_nrecord), dtype=np.int64)

    # Main
    for ix, peak_i in enumerate(peak_list):
        # reset buffers
        b_sumwf[:] = 0
        b_pulse[:] = 0
        b_index[:] = -1

        p = peaks[peak_i]
        channel_saturated = p['saturated_channel'] > 0

        for record_i in range(record_ranges[ix][0], record_ranges[ix][1]):
            r = records[record_i]
            r_slice, b_slice = strax.overlap_indices(
                r['time'] // dt, r['length'],
                p['time'] // dt, p['length'] * p['dt'] // dt)

            ch = r['channel']
            if channel_saturated[ch]:
                b_pulse[ch, slice(*b_slice)] += r['data'][slice(*r_slice)]
                b_index[ch, np.argmin(b_index[ch])] = record_i
            else:
                b_sumwf[slice(*b_slice)] += r['data'][slice(*r_slice)] \
                                            * to_pe[ch]

        _peak_saturation_correction_inner(
            channel_saturated, records, p,
            to_pe, b_sumwf, b_pulse, b_index,
            reference_length, min_reference_length)

        # Back track sum wf downsampling
        peaks[peak_i]['length'] = p['length'] * p['dt'] / dt
        peaks[peak_i]['dt'] = dt

    strax.sum_waveform(peaks, hitlets, records, rlinks, to_pe, n_top_channels, peak_list)
    return peak_list


@numba.jit(nopython=True, nogil=True, cache=True)
def _peak_saturation_correction_inner(channel_saturated, records, p,
                                      to_pe, b_sumwf, b_pulse, b_index,
                                      reference_length=100,
                                      min_reference_length=20,
                                      ):
    """Would add a third level loop in peak_saturation_correction
    Which is not ideal for numba, thus this function is written
    :param channel_saturated: (bool, n_channels)
    :param p: One peak/peaklet
    :param to_pe: adc to PE conversion (length should equal number of PMTs)
    :param b_sumwf, b_pulse, b_index: Filled buffers
    """
    dt = records['dt'][0]
    n_channels = len(channel_saturated)

    for ch in range(n_channels):
        if not channel_saturated[ch]:
            continue
        b = b_pulse[ch]
        r0 = records[b_index[ch][0]]

        # Define the reference region as reference_length before the first saturation point
        # unless there are not enough samples
        bl = np.inf
        for record_i in b_index[ch]:
            if record_i == -1:
                break
            bl = min(bl, records['baseline'][record_i])

        s0 = np.argmax(b >= np.int16(bl))
        ref = slice(max(0, s0 - reference_length), s0)

        if (b[ref] * to_pe[ch] > 1).sum() < min_reference_length:
            # the pulse is saturated, but there are not enough reference samples to get a good ratio
            # This actually distinguished between S1 and S2 and will only correct S2 signals
            continue
        if (b_sumwf[ref] > 1).sum() < min_reference_length:
            # the same condition applies to the waveform model
            continue
        if np.sum(b[ref]) * to_pe[ch] / np.sum(b_sumwf[ref]) > 1:
            # The pulse is saturated, but insufficient information is available in the other channels
            # to reliably reconstruct it
            continue

        scale = np.sum(b[ref]) / np.sum(b_sumwf[ref])

        # Loop over the record indices of the saturated channel (saved in b_index buffer)
        for record_i in b_index[ch]:
            if record_i == -1:
                break
            r = records[record_i]
            r_slice, b_slice = strax.overlap_indices(
                r['time'] // dt, r['length'],
                p['time'] // dt + s0, p['length'] * p['dt'] // dt - s0)

            if r_slice[1] == r_slice[0]:  # This record proceeds saturation
                continue
            b_slice = b_slice[0] + s0, b_slice[1] + s0

            # First is finding the highest point in the desaturated record
            # because we need to bit shift the whole record if it exceeds int16 range
            apax = scale * max(b_sumwf[slice(*b_slice)])

            if np.int32(apax) >= 2 ** 15:  # int16(2**15) is -2**15
                bshift = int(np.floor(np.log2(apax) - 14))

                tmp = r['data'].astype(np.int32)
                tmp[slice(*r_slice)] = b_sumwf[slice(*b_slice)] * scale

                r['area'] = np.sum(tmp)  # Auto covert to int64
                r['data'][:] = np.right_shift(tmp, bshift)
                r['amplitude_bit_shift'] += bshift
            else:
                r['data'][slice(*r_slice)] = b_sumwf[slice(*b_slice)] * scale
                r['area'] = np.sum(r['data'])


@numba.jit(nopython=True, nogil=True, cache=True)
def get_tight_coin(hit_max_times, hit_channel, peak_max_times, left, right,
                   channels=(0, 493)):
    """Calculates the tight coincidence based on PMT channels.

    Defined by number of hits within a specified time range of the
    the peak's maximum amplitude.
    Imitates tight_coincidence variable in pax:
    github.com/XENON1T/pax/blob/master/pax/plugins/peak_processing/BasicProperties.py

    :param hit_max_times: Time of the hit amplitude in ns.
    :param hit_channel: PMT channels of the hits
    :param peak_max_times: Time of the peaks maximum in ns.
    :param left: Left boundary in which we search for the tight
        coincidence in ns.
    :param right: Right boundary in which we search for the tight
        coincidence in ns.
    :param channel_range: (min/max) channel for the corresponding detector.

    :returns: n_coin_channel of length peaks containing the
        tight coincidence.
    """
    left_hit_i = 0
    n_coin_channel = np.zeros(len(peak_max_times), dtype=np.int16)
    start_ch, end_ch = channels
    channels_seen = np.zeros(end_ch - start_ch + 1, dtype=np.bool_)

    # loop over peaks
    for p_i, p_t in enumerate(peak_max_times):
        channels_seen[:] = 0
        # loop over hits starting from the last one we left at
        for left_hit_i in range(left_hit_i, len(hit_max_times)):

            # if the hit is in the window, its a tight coin
            d = hit_max_times[left_hit_i] - p_t
            if (-left <= d) & (d <= right):
                channels_seen[hit_channel[left_hit_i] - start_ch] = 1

            # stop the loop when we know we're outside the range
            if d > right:
                n_coin_channel[p_i] = np.sum(channels_seen)
                break

        # Add channel information in case there are no hits beyond
        # the last peak:
        n_coin_channel[p_i] = np.sum(channels_seen)

    return n_coin_channel


@numba.njit(cache=True, nogil=True)
def hit_max_sample(records, hits):
    """Return the index of the maximum sample for hits"""
    result = np.zeros(len(hits), dtype=np.int16)
    for i, h in enumerate(hits):
        r = records[h['record_i']]
        w = r['data'][h['left']:h['right']]
        result[i] = np.argmax(w)
    return result



import strax
import numpy as np

export, __all__ = strax.exporter()


@export
class nVETOEventsSync(strax.OverlapWindowPlugin):
    """
    Plugin which computes time stamps which are synchronized with the
    TPC. Uses delay set in the DAQ.
    """
    depends_on = ('events_nv', 'detector_time_offsets')
    delay_field_name = 'time_offset_nv'

    provides = 'events_sync_nv'
    save_when = strax.SaveWhen.EXPLICIT
    __version__ = '0.0.3'

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        dtype += [(('Time of the event synchronized according to the total digitizer delay.',
                    'time_sync'), np.int64),
                  (('Endtime of the event synchronized according to the total digitizer delay.',
                    'endtime_sync'), np.int64),
                  ]
        return dtype

    def get_window_size(self):
        # Ensure to have at least 12 offset-values from detector_time_offsets
        # to compute average time delay. Otherwise we may get unlucky with
        # our pacemaker (unlikely but could happen).
        return 120 * 10 ** 9

    def compute(self, events_nv, detector_time_offsets):
        delay = detector_time_offsets[self.delay_field_name]
        delay = np.median(delay[delay > 0])
        delay = delay.astype(np.int64)
        # Check if delay is >= 0 otherwise something went wrong with
        # the sync signal.
        assert delay >= 0, f'Missing the GPS sync signal for run {self.run_id}.'

        events_sync_nv = np.zeros(len(events_nv), self.dtype)
        events_sync_nv['time'] = events_nv['time']
        events_sync_nv['endtime'] = events_nv['endtime']
        events_sync_nv['time_sync'] = events_nv['time'] + delay
        events_sync_nv['endtime_sync'] = events_nv['endtime'] + delay
        return events_sync_nv



import numpy as np
import strax
import straxen
from warnings import warn
export, __all__ = strax.exporter()

@export
class EventTopBottomParams(strax.Plugin):
    """
    Pluging that computes timing characteristics of top and bottom waveforms 
    based on waveforms stored at event level for main/alt S1/S2
    """
    depends_on = ('event_info', 'event_waveform')
    provides = 'event_top_bottom_params'
    __version__ = '0.0.0'
    def infer_dtype(self):
        ## Populating data type information
        infoline = {'s1': 'main S1',
                    's2': 'main S2',
                    'alt_s1': 'alternative S1',
                    'alt_s2': 'alternative S2',
                   }
        ev_info_fields = self.deps['event_info'].dtype.fields
        dtype = []
        # populating APC and waveform samples
        self.ptypes = ['s1', 's2', 'alt_s1', 'alt_s2']
        self.arrs = ['top', 'bot']
        for type_ in self.ptypes:
            for arr_ in self.arrs:
                dtype+=[((f'Central time for {infoline[type_]} for {arr_} PMTs [ ns ]',
                          f'{type_}_center_time_{arr_}'),
                          ev_info_fields[f'{type_}_center_time'][0])]
                dtype+=[((f'Time between 10% and 50% area quantiles for {infoline[type_]} for {arr_} PMTs [ns]',
                          f'{type_}_rise_time_{arr_}'),
                          ev_info_fields[f'{type_}_rise_time'][0])]
                dtype+=[((f'Width (in ns) of the central 50% area of the peak for {arr_} PMTs of {infoline[type_]}',
                          f'{type_}_range_50p_area_{arr_}'),
                          ev_info_fields[f'{type_}_range_50p_area'][0])]
                dtype+=[((f'Width (in ns) of the central 90% area of the peak for {arr_} PMTs of {infoline[type_]}',
                          f'{type_}_range_90p_area_{arr_}'),
                          ev_info_fields[f'{type_}_range_90p_area'][0])]
            dtype+=[((f'Difference between center times of top and bottom arrays for {infoline[type_]} [ ns ]',
                      f'{type_}_center_time_diff_top_bot'),
                      ev_info_fields[f'{type_}_center_time'][0])]
        dtype += strax.time_fields
        return dtype

    def compute(self, events):
        result = np.zeros(events.shape, dtype=self.dtype)
        result['time'], result['endtime'] = events['time'], strax.endtime(events)
        peak_dtype = strax.peak_dtype(n_channels=straxen.n_tpc_pmts, digitize_top=False)
        for type_ in self.ptypes:
            for arr_ in self.arrs:
                # in order to reuse the same definitions as in other parts, we create "fake peaks"
                # based only on data from corresponding array
                fpeaks_ = np.zeros(events.shape[0], dtype=peak_dtype)
                if arr_ == 'top':
                    fpeaks_['data'] = events[f'{type_}_data_top']
                    fpeaks_['area'] = events[f'{type_}_area'] * events[f'{type_}_area_fraction_top']
                elif arr_ == 'bot':
                    fpeaks_['data'] = (events[f'{type_}_data'] - events[f'{type_}_data_top'])
                    fpeaks_['area'] = events[f'{type_}_area'] * (1.-events[f'{type_}_area_fraction_top'])
                elif arr_ == 'tot':
                    # This one is ony
                    fpeaks_['data'] = events[f'{type_}_data']
                    fpeaks_['area'] = events[f'{type_}_area']
                else:
                    raise RuntimeError(f'Received unknown array type : ' + arr_)
                fpeaks_['length'] = events[f'{type_}_length']
                fpeaks_['dt'] = events[f'{type_}_dt']
                # computing central times
                # note that here we ignore 1/2 sample length to be consistent with other definitions
                with np.errstate(divide='ignore', invalid='ignore'):
                    recalc_ctime = np.sum(fpeaks_['data'] * (np.arange(0, fpeaks_['data'].shape[1])), axis=1)
                    recalc_ctime /= fpeaks_['area']
                    recalc_ctime *= fpeaks_['dt']
                    recalc_ctime[~(fpeaks_['area'] > 0)] = 0.0
                # setting central times in the same way as inside peak processing
                mask = (fpeaks_['area'] > 0)
                result[f'{type_}_center_time_{arr_}'] = events[f'{type_}_time']
                result[f'{type_}_center_time_{arr_}'][mask] += recalc_ctime[mask].astype(int)
                # computing widths ##
                # zero or undefined area peaks should have nans
                strax.compute_widths(fpeaks_)
                result[f'{type_}_rise_time_{arr_}'][:] = np.nan
                result[f'{type_}_rise_time_{arr_}'][mask] = -fpeaks_['area_decile_from_midpoint'][mask][:, 1]
                result[f'{type_}_range_50p_area_{arr_}'][:] = np.nan
                result[f'{type_}_range_50p_area_{arr_}'][mask] = fpeaks_['width'][mask][:, 5]
                result[f'{type_}_range_90p_area_{arr_}'][:] = np.nan
                result[f'{type_}_range_90p_area_{arr_}'][mask] = fpeaks_['width'][mask][:, 9]
            # Difference between center times of top and bottom arrays
            result[f'{type_}_center_time_diff_top_bot'] = (result[f'{type_}_center_time_top'] -
                                                            result[f'{type_}_center_time_bot'])
        return result



from . import event_info
from .event_info import *

from . import pax_interface
from .pax_interface import *

from . import peak_positions
from .peak_positions import *

from . import x1t_cuts
from .x1t_cuts import *



import strax
import straxen
from straxen.plugins.peaklets.peaklets import drop_data_top_field

import numpy as np
import numba


export, __all__ = strax.exporter()


@export
class MergedS2s(strax.OverlapWindowPlugin):
    """
    Merge together peaklets if peak finding favours that they would
    form a single peak instead.
    """
    __version__ = '1.0.2'

    depends_on = ('peaklets', 'peaklet_classification', 'lone_hits')
    data_kind = 'merged_s2s'
    provides = 'merged_s2s'

    s2_merge_max_duration = straxen.URLConfig(
        default=50_000, infer_type=False,
        help="Do not merge peaklets at all if the result would be a peak "
             "longer than this [ns]")

    s2_merge_gap_thresholds = straxen.URLConfig(
        default=((1.7, 2.65e4), (4.0, 2.6e3), (5.0, 0.)),
        infer_type=False,
        help="Points to define maximum separation between peaklets to allow "
             "merging [ns] depending on log10 area of the merged peak\n"
             "where the gap size of the first point is the maximum gap to allow merging"
             "and the area of the last point is the maximum area to allow merging. "
             "The format is ((log10(area), max_gap), (..., ...), (..., ...))"
    )

    gain_model = straxen.URLConfig(
        infer_type=False,
        help='PMT gain model. Specify as '
             '(str(model_config), str(version), nT-->boolean')

    merge_without_s1 = straxen.URLConfig(
        default=True, infer_type=False,
        help="If true, S1s will be igored during the merging. "
             "It's now possible for a S1 to be inside a S2 post merging")

    n_top_pmts = straxen.URLConfig(
        type=int,
        help="Number of top TPC array PMTs")

    n_tpc_pmts = straxen.URLConfig(
        type=int,
        help='Number of TPC PMTs')

    sum_waveform_top_array = straxen.URLConfig(
        default=True,
        type=bool,
        help='Digitize the sum waveform of the top array separately'
    )

    merged_s2s_get_window_size_factor = straxen.URLConfig(
        default=5,
        type=int,
        track=False,
        help='Factor of the window size for the merged_s2s plugin'
    )

    def setup(self):
        self.to_pe = self.gain_model

    def infer_dtype(self):
        return strax.unpack_dtype(self.deps['peaklets'].dtype_for('peaklets'))

    def get_window_size(self):
        return self.merged_s2s_get_window_size_factor * (
            int(self.s2_merge_gap_thresholds[0][1]) + self.s2_merge_max_duration)

    def compute(self, peaklets, lone_hits):
        if self.merge_without_s1:
            peaklets = peaklets[peaklets['type'] != 1]

        if len(peaklets) <= 1:
            return np.zeros(0, dtype=self.dtype)

        gap_thresholds = self.s2_merge_gap_thresholds
        max_gap = gap_thresholds[0][1]
        max_area = 10 ** gap_thresholds[-1][0]

        if max_gap < 0:
            # Do not merge at all
            return np.zeros(0, dtype=self.dtype)

        if 'data_top' not in peaklets.dtype.names:
            peaklets_w_field = np.zeros(len(peaklets), dtype=strax.peak_dtype(n_channels=self.n_tpc_pmts, digitize_top=True))
            strax.copy_to_buffer(peaklets, peaklets_w_field, '_add_data_top_field')
            del peaklets
            peaklets = peaklets_w_field

        # Max gap and area should be set by the gap thresholds
        # to avoid contradictions
        start_merge_at, end_merge_at = self.get_merge_instructions(
            peaklets['time'], strax.endtime(peaklets),
            areas=peaklets['area'],
            types=peaklets['type'],
            gap_thresholds=gap_thresholds,
            max_duration=self.s2_merge_max_duration,
            max_gap=max_gap,
            max_area=max_area,
        )

        assert 'data_top' in peaklets.dtype.names

        merged_s2s = strax.merge_peaks(
            peaklets,
            start_merge_at, end_merge_at,
            max_buffer=int(self.s2_merge_max_duration // np.gcd.reduce(peaklets['dt'])),
        )
        merged_s2s['type'] = 2

        # Updated time and length of lone_hits and sort again:
        lh = np.copy(lone_hits)
        del lone_hits
        lh_time_shift = (lh['left'] - lh['left_integration']) * lh['dt']
        lh['time'] = lh['time'] - lh_time_shift
        lh['length'] = (lh['right_integration'] - lh['left_integration'])
        lh = strax.sort_by_time(lh)

        # If sum_waveform_top_array is false, don't digitize the top array
        n_top_pmts_if_digitize_top = self.n_top_pmts if self.sum_waveform_top_array else -1
        strax.add_lone_hits(merged_s2s, lh, self.to_pe, n_top_channels=n_top_pmts_if_digitize_top)

        strax.compute_widths(merged_s2s)

        if n_top_pmts_if_digitize_top <= 0:
            merged_s2s = drop_data_top_field(merged_s2s, self.dtype, '_drop_top_merged_s2s')
        return merged_s2s

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def get_merge_instructions(
            peaklet_starts, peaklet_ends, areas, types,
            gap_thresholds, max_duration, max_gap, max_area,
            sort_kind='mergesort'):
        """
        Finding the group of peaklets to merge. To do this start with the
        smallest gaps and keep merging until the new, merged S2 has such a
        large area or gap to adjacent peaks that merging is not required
        anymore.
        see https://github.com/XENONnT/straxen/pull/548 and https://github.com/XENONnT/straxen/pull/568

        :returns: list of the first index of peaklet to be merged and
        list of the exclusive last index of peaklet to be merged
        """

        peaklet_gaps = peaklet_starts[1:] - peaklet_ends[:-1]
        peaklet_start_index = np.arange(len(peaklet_starts))
        peaklet_end_index = np.arange(len(peaklet_starts))

        for gap_i in np.argsort(peaklet_gaps, kind=sort_kind):
            start_idx = peaklet_start_index[gap_i]
            inclusive_end_idx = peaklet_end_index[gap_i + 1]
            sum_area = np.sum(areas[start_idx:inclusive_end_idx + 1])
            this_gap = peaklet_gaps[gap_i]

            if inclusive_end_idx < start_idx:
                raise ValueError('Something went wrong, left is bigger then right?!')

            if this_gap > max_gap:
                break
            if sum_area > max_area:
                # For very large S2s, we assume that natural breaks is taking care
                continue
            if (sum_area > 0) and (
                    this_gap > merge_s2_threshold(np.log10(sum_area),
                                                  gap_thresholds)):
                # The merged peak would be too large
                continue

            peak_duration = (peaklet_ends[inclusive_end_idx] - peaklet_starts[start_idx])
            if peak_duration >= max_duration:
                continue

            # Merge gap in other words this means p @ gap_i and p @gap_i + 1 share the same
            # start, end and area:
            peaklet_start_index[start_idx:inclusive_end_idx + 1] = peaklet_start_index[start_idx]
            peaklet_end_index[start_idx:inclusive_end_idx + 1] = peaklet_end_index[inclusive_end_idx]

        start_merge_at = np.unique(peaklet_start_index)
        end_merge_at = np.unique(peaklet_end_index)
        if not len(start_merge_at) == len(end_merge_at):
            raise ValueError('inconsistent start and end merge instructions')

        merge_start, merge_stop_exclusive = _filter_s1_starts(
            start_merge_at, types, end_merge_at)

        return merge_start, merge_stop_exclusive


@numba.njit(cache=True, nogil=True)
def _filter_s1_starts(start_merge_at, types, end_merge_at):
    for start_merge_idx, _ in enumerate(start_merge_at):
        while types[start_merge_at[start_merge_idx]] != 2:
            if end_merge_at[start_merge_idx] - start_merge_at[start_merge_idx] <= 1:
                break
            start_merge_at[start_merge_idx] += 1

    start_merge_with_s2 = types[start_merge_at] == 2
    merges_at_least_two_peaks = end_merge_at - start_merge_at >= 1

    keep_merges = start_merge_with_s2 & merges_at_least_two_peaks
    return start_merge_at[keep_merges], end_merge_at[keep_merges] + 1


@numba.njit(cache=True, nogil=True)
def merge_s2_threshold(log_area, gap_thresholds):
    """Return gap threshold for log_area of the merged S2
    with linear interpolation given the points in gap_thresholds
    :param log_area: Log 10 area of the merged S2
    :param gap_thresholds: tuple (n, 2) of fix points for interpolation
    """
    for i, (a1, g1) in enumerate(gap_thresholds):
        if log_area < a1:
            if i == 0:
                return g1
            a0, g0 = gap_thresholds[i - 1]
            return (log_area - a0) * (g1 - g0) / (a1 - a0) + g0
    return gap_thresholds[-1][1]



import typing
import numpy as np
import strax
import straxen

from straxen.plugins.aqmon_hits.aqmon_hits import AqmonChannels

export, __all__ = strax.exporter()


# ### Veto hardware ###:
# V1495 busy veto module:
# Generates a 25 ns NIM pulse whenever a veto begins and a 25 ns NIM signal when it ends.
# A new start signal can occur only after the previous busy instance ended.
# 1ms (1e6 ns) - minimum busy veto length, or until the board clears its memory

# DDC10 High Energy Veto:
# 10ms (1e7 ns) - fixed HE veto length in XENON1T DDC10,
# in XENONnT it will be calibrated based on the length of large S2 SE tails
# The start/stop signals for the HEV are generated by the V1495 board


@export
class VetoIntervals(strax.OverlapWindowPlugin):
    """ Find pairs of veto start and veto stop signals and the veto
    duration between them:
     - busy_*  <= V1495 busy veto for tpc channels
     - busy_he_*    <= V1495 busy veto for high energy tpc channels
     - hev_*   <= DDC10 hardware high energy veto
     - straxen_deadtime <= special case of deadtime introduced by the
       DAQReader-plugin
    """
    __version__ = '1.1.1'
    depends_on = 'aqmon_hits'
    provides = 'veto_intervals'
    data_kind = 'veto_intervals'

    # This option is just showing where the OverlapWindowPlugin fails.
    # We need to buffer the entire run in order not to run into chunking
    # issues. A better solution would be using
    #   github.com/AxFoundation/strax/pull/654
    max_veto_window = straxen.URLConfig(
        default=int(7.2e12),
        track=True,
        type=int,
        help='Maximum separation between veto stop and start pulses [ns]. '
             'Set to be >> than the max duration of the run to be able to '
             'fully store one run into buffer since aqmon-hits are not '
             'sorted by endtime'
    )

    def infer_dtype(self):
        dtype = [(('veto interval [ns]', 'veto_interval'), np.int64),
                 (('veto signal type', 'veto_type'), np.str_('U30'))]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.veto_names = ['busy_', 'busy_he_', 'hev_']
        self.channel_map = {aq_ch.name.lower(): int(aq_ch)
                            for aq_ch in AqmonChannels}

    def get_window_size(self):
        # Give a very wide window
        return self.max_veto_window

    def compute(self, aqmon_hits, start, end):
        # Allocate a nice big buffer and throw away the part we don't need later
        result = np.zeros(len(aqmon_hits) * len(self.veto_names), self.dtype)
        vetos_seen = 0

        for veto_name in self.veto_names:
            veto_hits_start = channel_select(aqmon_hits, self.channel_map[veto_name + 'start'])
            veto_hits_stop = channel_select(aqmon_hits, self.channel_map[veto_name + 'stop'])

            veto_hits_start, veto_hits_stop = self.handle_starts_and_stops_outside_of_run(
                veto_hits_start=veto_hits_start,
                veto_hits_stop=veto_hits_stop,
                chunk_start=start,
                chunk_end=end,
                veto_name=veto_name,
            )
            n_vetos = len(veto_hits_start)

            result["time"][vetos_seen:vetos_seen + n_vetos] = veto_hits_start['time']
            result["endtime"][vetos_seen:vetos_seen + n_vetos] = veto_hits_stop['time']
            result["veto_type"][vetos_seen:vetos_seen + n_vetos] = veto_name + 'veto'

            vetos_seen += n_vetos

        # Straxen deadtime is special, it's a start and stop with no data
        # but already an interval so easily used here
        artificial_deadtime = aqmon_hits[(aqmon_hits['channel'] ==
                                          AqmonChannels.ARTIFICIAL_DEADTIME)]
        n_artificial = len(artificial_deadtime)

        if n_artificial:
            result[vetos_seen:n_artificial]['time'] = artificial_deadtime['time']
            result[vetos_seen:n_artificial]['endtime'] = strax.endtime(artificial_deadtime)
            result[vetos_seen:n_artificial]['veto_type'] = 'straxen_deadtime_veto'
            vetos_seen += n_artificial

        result = result[:vetos_seen]
        result['veto_interval'] = result['endtime'] - result['time']
        sort = np.argsort(result['time'])
        result = result[sort]
        return result

    def handle_starts_and_stops_outside_of_run(
            self,
            veto_hits_start: np.ndarray,
            veto_hits_stop: np.ndarray,
            chunk_start: int,
            chunk_end: int,
            veto_name: str,
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """
        We might be missing one start or one stop at the end of the run,
        set it to the chunk endtime if this is the case
        """
        # Just for traceback info that we declare this here
        extra_start = []
        extra_stop = []
        missing_a_final_stop = (
                len(veto_hits_start)
                and len(veto_hits_stop)
                and veto_hits_start[-1]['time'] > veto_hits_stop['time'][-1])
        missing_a_final_stop = (
                missing_a_final_stop
                or (len(veto_hits_start) and not len(veto_hits_stop)
                    )
        )
        if missing_a_final_stop:
            # There is one *start* of the //end// of the run -> the
            # **stop** is missing (because it's outside of the run),
            # let's add one **stop** at the //end// of this chunk
            extra_stop = self.fake_hit(chunk_end)
            veto_hits_stop = np.concatenate([veto_hits_stop, extra_stop])
        if len(veto_hits_stop) - len(veto_hits_start) == 1:
            # There is one *stop* of the //beginning// of the run
            # -> the **start** is missing (because it's from before
            # starting the run), # let's add one **start** at the
            # //beginning// of this chunk
            extra_start = self.fake_hit(chunk_start)
            veto_hits_start = np.concatenate([extra_start, veto_hits_start])

        something_is_wrong = len(veto_hits_start) != len(veto_hits_stop)

        message = (f'Got inconsistent number of {veto_name} starts '
                   f'{len(veto_hits_start)}) / stops ({len(veto_hits_stop)}).')
        if len(extra_start):
            message += ' Despite the fact that we inserted one extra start at the beginning of the run.'  # noqa
        elif len(extra_stop):
            message += ' Despite the fact that we inserted one extra stop at the end of the run.'  # noqa
        if something_is_wrong:
            raise ValueError(message)

        if np.any(veto_hits_start['time'] > veto_hits_stop['time']):
            raise ValueError('Found veto\'s starting before the previous stopped')

        return veto_hits_start, veto_hits_stop

    @staticmethod
    def fake_hit(start, dt=1, length=1):
        hit = np.zeros(1, strax.hit_dtype)
        hit['time'] = start
        hit['dt'] = dt
        hit['length'] = length
        return hit


# Don't use @numba since numba doesn't like masking arrays, use numpy
def channel_select(rr, ch):
    """Return data from start/stop veto channel in the acquisition monitor (AM)"""
    return rr[rr['channel'] == ch]



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakAmbience(strax.OverlapWindowPlugin):
    """
    Calculate Ambience of peaks.
    Features are the number of lonehits, small S0, S1, S2 in a time window before peaks,
    and the number of small S2 in circle near the S2 peak in a time window.
    References:
        * v0.0.7 reference: xenon:xenonnt:ac:prediction:shadow_ambience
    """
    __version__ = '0.0.8'
    depends_on = ('lone_hits', 'peak_basics', 'peak_positions')
    provides = 'peak_ambience'
    data_kind = 'peaks'
    save_when = strax.SaveWhen.EXPLICIT

    ambience_time_window_backward = straxen.URLConfig(
        default=int(2e6),
        type=int,
        track=True,
        help='Search for ambience in this time window [ns]'
    )

    ambience_divide_t = straxen.URLConfig(
        default=False,
        type=bool,
        track=True,
        help='Whether to divide area by time difference of ambience creating peak to current peak'
    )

    ambience_divide_r = straxen.URLConfig(
        default=False,
        type=bool,
        track=True,
        help='Whether to divide area by radial distance of ambience creating peak to current peak'
    )

    ambient_radius = straxen.URLConfig(
        default=6.7,
        type=float,
        track=True,
        help='Search for ambience in this radius [cm]'
    )

    ambience_area_parameters = straxen.URLConfig(
        default=(5, 60, 60),
        type=(list, tuple),
        track=True,
        help='The upper limit of S0, S1, S2 area to be counted'
    )

    def get_window_size(self):
        return 10 * self.ambience_time_window_backward

    @property
    def origin_dtype(self):
        return ['lh_before', 's0_before', 's1_before', 's2_before', 's2_near']

    def infer_dtype(self):
        dtype = []
        for ambience in self.origin_dtype:
            dtype += [((f"Number of small {' '.join(ambience.split('_'))} a peak",
                        f'n_{ambience}'), np.int16),
                      ((f"Area sum of small {' '.join(ambience.split('_'))} a peak",
                        f's_{ambience}'), np.float32)]
        dtype += strax.time_fields
        return dtype

    def compute(self, lone_hits, peaks):
        argsort = np.argsort(peaks['center_time'], kind='mergesort')
        _peaks = np.sort(peaks, order='center_time')
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_ambience(lone_hits, peaks, _peaks))
        return result

    def compute_ambience(self, lone_hits, peaks, current_peak):
        # 1. Initialization
        result = np.zeros(len(current_peak), self.dtype)

        # 2. Define time window for each peak, we will find small peaks & lone hits within these time windows
        roi = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi['time'] = current_peak['center_time'] - self.ambience_time_window_backward
        roi['endtime'] = current_peak['center_time']

        # 3. Calculate number and area sum of lonehits before a peak
        touching_windows = strax.touching_windows(lone_hits, roi)
        # Calculating ambience
        self.lonehits_ambience(current_peak,
                               lone_hits,
                               touching_windows,
                               result['n_lh_before'],
                               result['s_lh_before'],
                               self.ambience_divide_t)

        # 4. Calculate number and area sum of small S0, S1, S2 before a peak
        radius = -1
        for stype, area in zip([0, 1, 2],
                               self.ambience_area_parameters):
            mask_pre = (peaks['type'] == stype) & (peaks['area'] < area)
            touching_windows = strax.touching_windows(peaks[mask_pre], roi)
            # Calculating ambience
            self.peaks_ambience(current_peak,
                                peaks[mask_pre],
                                touching_windows,
                                radius,
                                result[f'n_s{stype}_before'],
                                result[f's_s{stype}_before'],
                                self.ambience_divide_t,
                                self.ambience_divide_r)

        # 5. Calculate number and area sum of small S2 near(in (x,y) space) a S2 peak
        mask_pre = (peaks['type'] == 2) & (peaks['area'] < self.ambience_area_parameters[2])
        touching_windows = strax.touching_windows(peaks[mask_pre], roi)
        # Calculating ambience
        self.peaks_ambience(current_peak,
                            peaks[mask_pre],
                            touching_windows,
                            self.ambient_radius,
                            result['n_s2_near'],
                            result['s_s2_near'],
                            self.ambience_divide_t,
                            self.ambience_divide_r)

        # 6. Set time and endtime for peaks
        result['time'] = current_peak['time']
        result['endtime'] = strax.endtime(current_peak)
        return result

    @staticmethod
    @numba.njit
    def lonehits_ambience(peaks, pre_hits,
                          touching_windows, num_array,
                          sum_array, ambience_divide_t):
        # Function to find lonehits before a peak
        # creating_hit is the lonehit creating ambience
        # suspicious_peak is the suspicious peak in the ambience created by creating_hit
        for p_i, suspicious_peak in enumerate(peaks):
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                creating_hit = pre_hits[idx]
                dt = suspicious_peak['center_time'] - creating_hit['time']
                if (dt <= 0) or (creating_hit['area'] <= 0):
                    continue
                num_array[p_i] += 1
                # Sometimes we may interested in sum of area / dt
                if ambience_divide_t:
                    sum_array[p_i] += creating_hit['area'] / dt
                else:
                    sum_array[p_i] += creating_hit['area']

    @staticmethod
    @numba.njit
    def peaks_ambience(peaks, pre_peaks,
                       touching_windows, ambient_radius,
                       num_array, sum_array,
                       ambience_divide_t, ambience_divide_r):
        # Function to find S0, S1, S2 before or near a peak
        # creating_peak is the peak creating ambience
        # suspicious_peak is the suspicious peak in the ambience created by creating_peak
        for p_i, suspicious_peak in enumerate(peaks):
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                creating_peak = pre_peaks[idx]
                r = distance_in_xy(suspicious_peak, creating_peak)
                dt = suspicious_peak['center_time'] - creating_peak['center_time']
                if dt <= 0:
                    continue
                if (ambient_radius < 0) or (r <= ambient_radius):
                    num_array[p_i] += 1
                    # Sometimes we may interested in sum of area / dt
                    if ambience_divide_t:
                        sum_array[p_i] += creating_peak['area'] / dt
                    else:
                        sum_array[p_i] += creating_peak['area']
                    # Sometimes we may interested in sum of area / r^2
                    if ambience_divide_r and ambient_radius > 0:
                        sum_array[p_i] /= r ** 2


@numba.njit
def distance_in_xy(peak_a, peak_b):
    """Distance between S2s in (x,y)"""
    return np.sqrt((peak_a['x'] - peak_b['x']) ** 2 +
                   (peak_a['y'] - peak_b['y']) ** 2)


@numba.njit
def _quick_assign(indices, results, inputs):
    for i, r in zip(indices, inputs):
        results[i] = r



import numpy as np
import strax
import straxen
from straxen.common import get_resource, rotate_perp_wires
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class CorrectedAreas(strax.Plugin):
    """
    Plugin which applies light collection efficiency maps and electron
    life time to the data.
    Computes the cS1/cS2 for the main/alternative S1/S2 as well as the
    corrected life time.
    Note:
        Please be aware that for both, the main and alternative S1, the
        area is corrected according to the xy-position of the main S2.
        There are now 3 components of cS2s: cs2_top, cS2_bottom and cs2.
        cs2_top and cs2_bottom are corrected by the corresponding maps,
        and cs2 is the sum of the two.
    """
    __version__ = '0.5.1'

    depends_on = ['event_basics', 'event_positions']

    # Descriptor configs
    elife = straxen.URLConfig(
        default='cmt://elife?version=ONLINE&run_id=plugin.run_id',
        help='electron lifetime in [ns]')

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO,
        help="default reconstruction algorithm that provides (x,y)"
    )
    s1_xyz_map = straxen.URLConfig(
        default='itp_map://resource://cmt://format://'
                's1_xyz_map_{algo}?version=ONLINE&run_id=plugin.run_id'
                '&fmt=json&algo=plugin.default_reconstruction_algorithm',
        cache=True)
    s2_xy_map = straxen.URLConfig(
        default='itp_map://resource://cmt://format://'
                's2_xy_map_{algo}?version=ONLINE&run_id=plugin.run_id'
                '&fmt=json&algo=plugin.default_reconstruction_algorithm',
        cache=True)

    # average SE gain for a given time period. default to the value of this run in ONLINE model
    # thus, by default, there will be no time-dependent correction according to se gain
    avg_se_gain = straxen.URLConfig(
        default='cmt://avg_se_gain?version=ONLINE&run_id=plugin.run_id',
        help='Nominal single electron (SE) gain in PE / electron extracted. '
             'Data will be corrected to this value')

    # se gain for this run, allowing for using CMT. default to online
    se_gain = straxen.URLConfig(
        default='cmt://se_gain?version=ONLINE&run_id=plugin.run_id',
        help='Actual SE gain for a given run (allows for time dependence)')

    # relative extraction efficiency which can change with time and modeled by CMT.
    rel_extraction_eff = straxen.URLConfig(
        default='cmt://rel_extraction_eff?version=ONLINE&run_id=plugin.run_id',
        help='Relative extraction efficiency for this run (allows for time dependence)')

    # relative light yield
    # defaults to no correction
    rel_light_yield = straxen.URLConfig(
        default='cmt://relative_light_yield?version=ONLINE&run_id=plugin.run_id',
        help='Relative light yield (allows for time dependence)'
    )

    region_linear = straxen.URLConfig(
        default=28,
        help='linear cut (cm) for ab region, check out the note https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_2_region_se_correction'
    )

    region_circular = straxen.URLConfig(
        default=60,
        help='circular cut (cm) for ab region, check out the note https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_2_region_se_correction'
    )

    # cS2 AFT correction due to photon ionization
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:zihao:sr1_s2aft_photonionization_correction
    cs2_bottom_top_ratio_correction = straxen.URLConfig(
        default=1,
        help='Scaling factor for cS2 AFT correction due to photon ionization'
    )

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields

        for peak_type, peak_name in zip(['', 'alt_'], ['main', 'alternate']):
            # Only apply 
            dtype += [
                (f'{peak_type}cs1', np.float32, f'Corrected area of {peak_name} S1 [PE]'),
                (
                    f'{peak_type}cs1_wo_timecorr', np.float32,
                    f'Corrected area of {peak_name} S1 (before LY correction) [PE]',
                ),
            ]
            names = ['_wo_timecorr', '_wo_picorr', '_wo_elifecorr', '']
            descriptions = ['S2 xy', 'SEG/EE', 'photon ionization', 'elife']
            for i, name in enumerate(names):
                if i == len(names) - 1:
                    description = ''
                elif i == 0:
                    # special treatment for wo_timecorr, apply elife correction
                    description = ' (before ' + ' + '.join(descriptions[i + 1:-1])
                    description += ', after ' + ' + '.join(
                        descriptions[:i + 1] + descriptions[-1:]) + ')'
                else:
                    description = ' (before ' + ' + '.join(descriptions[i + 1:])
                    description += ', after ' + ' + '.join(descriptions[:i + 1]) + ')'
                dtype += [
                    (
                        f'{peak_type}cs2{name}', np.float32,
                        f'Corrected area of {peak_name} S2{description} [PE]',
                    ),
                    (
                        f'{peak_type}cs2_area_fraction_top{name}', np.float32,
                        f'Fraction of area seen by the top PMT array for corrected '
                        f'{peak_name} S2{description}',
                    ),
                ]
        return dtype

    def ab_region(self, x, y):
        new_x, new_y = rotate_perp_wires(x, y)
        cond = new_x < self.region_linear
        cond &= new_x > -self.region_linear
        cond &= new_x ** 2 + new_y ** 2 < self.region_circular ** 2
        return cond

    def cd_region(self, x, y):
        return ~self.ab_region(x, y)

    def s2_map_names(self):

        # S2 top and bottom are corrected separately, and cS2 total is the sum of the two
        # figure out the map name
        if len(self.s2_xy_map.map_names) > 1:
            s2_top_map_name = "map_top"
            s2_bottom_map_name = "map_bottom"
        else:
            s2_top_map_name = "map"
            s2_bottom_map_name = "map"

        return s2_top_map_name, s2_bottom_map_name

    def seg_ee_correction_preparation(self):
        """Get single electron gain and extraction efficiency options"""
        self.regions = {'ab': self.ab_region, 'cd': self.cd_region}

        # setup SEG and EE corrections
        # if they are dicts, we just leave them as is
        # if they are not, we assume they are floats and
        # create a dict with the same correction in each region
        if isinstance(self.se_gain, dict):
            seg = self.se_gain
        else:
            seg = {key: self.se_gain for key in self.regions}

        if isinstance(self.avg_se_gain, dict):
            avg_seg = self.avg_se_gain
        else:
            avg_seg = {key: self.avg_se_gain for key in self.regions}

        if isinstance(self.rel_extraction_eff, dict):
            ee = self.rel_extraction_eff
        else:
            ee = {key: self.rel_extraction_eff for key in self.regions}
        
        return seg, avg_seg, ee

    def compute(self, events):
        result = np.zeros(len(events), self.dtype)
        result['time'] = events['time']
        result['endtime'] = events['endtime']

        # S1 corrections depend on the actual corrected event position.
        # We use this also for the alternate S1; for e.g. Kr this is
        # fine as the S1 correction varies slowly.
        event_positions = np.vstack([events['x'], events['y'], events['z']]).T

        for peak_type in ["", "alt_"]:
            result[f"{peak_type}cs1_wo_timecorr"] = (
                events[f'{peak_type}s1_area'] / self.s1_xyz_map(event_positions))
            result[f"{peak_type}cs1"] = (
                result[f"{peak_type}cs1_wo_timecorr"] / self.rel_light_yield)

        # S2 corrections
        s2_top_map_name, s2_bottom_map_name = self.s2_map_names()
        seg, avg_seg, ee = self.seg_ee_correction_preparation()

        # now can start doing corrections
        for peak_type in ["", "alt_"]:
            # S2(x,y) corrections use the observed S2 positions
            s2_positions = np.vstack([events[f'{peak_type}s2_x'], events[f'{peak_type}s2_y']]).T

            # corrected S2 with S2(x,y) map only, i.e. no elife correction
            # this is for S2-only events which don't have drift time info
            s2_xy_top = self.s2_xy_map(s2_positions, map_name=s2_top_map_name)
            cs2_top_xycorr = (
                events[f'{peak_type}s2_area'] * events[f'{peak_type}s2_area_fraction_top']
                / s2_xy_top)
            s2_xy_bottom = self.s2_xy_map(s2_positions, map_name=s2_bottom_map_name)
            cs2_bottom_xycorr = (
                events[f'{peak_type}s2_area'] * (1 - events[f'{peak_type}s2_area_fraction_top'])
                / s2_xy_bottom)

            # collect electron lifetime correction
            # for electron lifetime corrections to the S2s,
            # use drift time computed using the main S1.
            el_string = peak_type + "s2_interaction_" if peak_type == "alt_" else peak_type
            elife_correction = np.exp(events[f'{el_string}drift_time'] / self.elife)

            # collect SEG and EE corrections
            seg_ee_corr = np.zeros(len(events))
            for partition, func in self.regions.items():
                # partitioned SEG and EE
                partition_mask = func(events[f'{peak_type}s2_x'], events[f'{peak_type}s2_y'])
                # correct for SEG and EE
                seg_ee_corr[partition_mask] = seg[partition] / avg_seg[partition] * ee[partition]

            # apply S2 xy correction
            cs2_xycorr = cs2_top_xycorr + cs2_bottom_xycorr
            result[f"{peak_type}cs2_wo_timecorr"] = cs2_xycorr * elife_correction
            result[f"{peak_type}cs2_area_fraction_top_wo_timecorr"] = cs2_top_xycorr / cs2_xycorr

            # apply SEG and EE correction
            cs2_top_wo_picorr = cs2_top_xycorr / seg_ee_corr
            cs2_bottom_wo_picorr = cs2_bottom_xycorr / seg_ee_corr
            result[f"{peak_type}cs2_wo_picorr"] = cs2_top_wo_picorr + cs2_bottom_wo_picorr
            result[f"{peak_type}cs2_area_fraction_top_wo_picorr"] = (
                cs2_top_wo_picorr / result[f"{peak_type}cs2_wo_picorr"])

            # apply photon ionization intensity and cS2 AFT correction (see #1247)
            # cS2 bottom should be corrected by photon ionization, but not cS2 top
            cs2_top_wo_elifecorr = cs2_top_wo_picorr
            cs2_bottom_wo_elifecorr = (
                cs2_bottom_wo_picorr * self.cs2_bottom_top_ratio_correction)
            result[f"{peak_type}cs2_wo_elifecorr"] = cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr
            result[f"{peak_type}cs2_area_fraction_top_wo_elifecorr"] = (
                cs2_top_wo_elifecorr / result[f"{peak_type}cs2_wo_elifecorr"])

            # apply electron lifetime correction
            result[f"{peak_type}cs2"] = result[f"{peak_type}cs2_wo_elifecorr"] * elife_correction
            result[f"{peak_type}cs2_area_fraction_top"] = (
                result[f"{peak_type}cs2_area_fraction_top_wo_elifecorr"])
        return result
